# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import datetime
import sys
from tempfile import TemporaryDirectory

import pytest

from flow import FlowProject
from flow.directives import (
    _GET_EXECUTABLE,
    _MEMORY,
    _NGPU,
    _NP,
    _NRANKS,
    _OMP_NUM_THREADS,
    _PROCESSOR_FRACTION,
    _WALLTIME,
    _Directive,
    _Directives,
    _no_aggregation,
)
from flow.errors import DirectivesError


@pytest.fixture()
def available_directives_list():
    return [
        _NP,
        _NRANKS,
        _NGPU,
        _OMP_NUM_THREADS,
        _GET_EXECUTABLE(),
        _WALLTIME,
        _MEMORY,
        _PROCESSOR_FRACTION,
    ]


@pytest.fixture()
def directives(available_directives_list):
    return _Directives(available_directives_list)


@pytest.fixture()
def product_directive():
    def val(v):
        if v < 10:
            raise ValueError("Price cannot be less than 10 units")
        return v

    def finalize(value, dict):
        discount = dict.get("discount", 0)
        free = dict.get("free", False)
        value = value - discount
        if value < 0 or free:
            return 0
        return value

    product = _Directive(
        name="product",
        validator=val,
        default=10,
        serial=_no_aggregation,
        parallel=_no_aggregation,
        finalize=finalize,
    )
    return product


@pytest.fixture()
def non_default_directive_values():
    return [
        {
            "np": 1,
            "ngpu": 10,
            "nranks": 5,
            "omp_num_threads": 20,
            "executable": "Non Default Path",
            "walltime": 64.0,
            "memory": 32,
            "processor_fraction": 0.5,
        },
        {
            "np": 4,
            "ngpu": 1,
            "nranks": 0,
            "omp_num_threads": 10,
            "executable": "PathFinder",
            "walltime": 20.0,
            "memory": 16,
            "processor_fraction": 0.5,
        },
    ]


class TestItems:
    """Tests for _Directive class."""

    def test_default(self):
        assert _NP._default == 1
        assert _NGPU._default == 0
        assert _NRANKS._default == 0
        assert _OMP_NUM_THREADS._default == 0
        assert _GET_EXECUTABLE()._default == sys.executable
        assert _WALLTIME._default is None
        assert _MEMORY._default is None
        assert _PROCESSOR_FRACTION._default == 1.0

    def test_invalid_values(self, available_directives_list):
        invalid_values = {
            "np": [-1, "foo", {}, None],
            "ngpu": [-1, "foo", {}, None],
            "nranks": [-1, "foo", {}, None],
            "omp_num_threads": [-1, "foo", {}, None],
            "walltime": [-1, "foo", {}],
            "memory": [-1, "foo", {}],
            "processor_fraction": [-0.5, 2.5, "foo", {}, None],
        }

        for directive in available_directives_list:
            if directive._name == "executable":
                # Executable expect a string, if not found, then it tries to convert
                # it into a string and becomes successful almost every time.
                # Hence skipping Executable.
                continue
            for i, value in enumerate(invalid_values[directive._name]):
                with pytest.raises((ValueError, TypeError)):
                    directive._validator(value)

    def test_defaults_are_valid(self, available_directives_list):
        for directive in available_directives_list:
            directive._validator(directive._default)

    def test_serial(self):
        assert _NP._serial(4, 2) == 4
        assert _NRANKS._serial(4, 2) == 4
        assert _NGPU._serial(4, 2) == 4
        assert _OMP_NUM_THREADS._serial(4, 2) == 4
        assert _GET_EXECUTABLE()._serial("Path1", "Path2") == "Path1"
        assert _WALLTIME._serial(4, 2) == 6
        assert _WALLTIME._serial(4, None) == 4
        assert _WALLTIME._serial(None, 4) == 4
        assert _WALLTIME._serial(None, None) is None
        assert _MEMORY._serial(4, 2) == 4
        assert _MEMORY._serial(4, None) == 4
        assert _MEMORY._serial(None, 4) == 4
        assert _MEMORY._serial(None, None) is None
        assert _PROCESSOR_FRACTION._serial(0.4, 0.2) == 0.4

    def test_parallel(self):
        assert _NP._parallel(4, 2) == 6
        assert _NRANKS._parallel(4, 2) == 6
        assert _NGPU._parallel(4, 2) == 6
        assert _OMP_NUM_THREADS._parallel(4, 2) == 6
        assert _GET_EXECUTABLE()._parallel("Path1", "Path2") == "Path1"
        assert _WALLTIME._parallel(4, 2) == 4
        assert _WALLTIME._parallel(4, None) == 4
        assert _WALLTIME._parallel(None, 4) == 4
        assert _WALLTIME._parallel(None, None) is None
        assert _MEMORY._parallel(4, 2) == 6
        assert _MEMORY._parallel(4, None) == 4
        assert _MEMORY._parallel(None, 4) == 4
        assert _MEMORY._parallel(None, None) is None
        assert _PROCESSOR_FRACTION._parallel(0.4, 0.2) == 0.4

    def test_finalize(self):
        dict_directives = {
            "nranks": _NRANKS._default,
            "omp_num_threads": _OMP_NUM_THREADS._default,
        }
        assert _NP._finalize(2, dict_directives) == 2
        dict_directives["nranks"] = 2
        dict_directives["omp_num_threads"] = 4
        assert _NP._finalize(2, dict_directives) == 2
        assert _NP._finalize(1, dict_directives) == 8

    def test_manual_item_default(self, product_directive):
        assert product_directive._default == 10

    def test_manual_item_validation(self, product_directive):
        val = product_directive._validator(product_directive._default)
        assert product_directive._default == val
        assert product_directive._validator(20) == 20
        with pytest.raises(ValueError):
            product_directive._validator(0)

    def test_manual_item_serial(self, product_directive):
        product_directive._serial(10, 20) == 10
        product_directive._serial(20, 10) == 20

    def test_manual_item_parallel(self, product_directive):
        product_directive._parallel(10, 20) == 10

    def test_manual_item_finalize(self, product_directive):
        asset_dict = {"free": False, "discount": 5}
        assert product_directive._finalize(50, asset_dict) == 45
        asset_dict["free"] = True
        assert product_directive._finalize(50, asset_dict) == 0


class TestDirectives:
    """Tests for _Directives Class."""

    def test_get_directive(self, directives, available_directives_list):
        for item in available_directives_list:
            assert directives[item._name] == item._default

    def test_add_directive(self, available_directives_list):
        directives = _Directives(available_directives_list[:-1])
        directives._add_directive(_PROCESSOR_FRACTION)
        assert directives[_PROCESSOR_FRACTION._name] == _PROCESSOR_FRACTION._default
        with pytest.raises(TypeError):
            directives._add_directive("Test")
        with pytest.raises(ValueError):
            directives._add_directive(_PROCESSOR_FRACTION)

    def test_set_defined_directive(self, directives):
        directives._set_defined_directive(_NP._name, 10)
        assert directives[_NP._name] == 10

    def test_set_defined_directive_invalid(self, directives):
        with pytest.raises(ValueError):
            directives._set_defined_directive(_NP._name, 0)

    def test_set_undefined_directive(self, directives):
        with pytest.raises(DirectivesError):
            directives._set_defined_directive("test", 0)

    def test_set_directives_item(self, directives):
        directives["test"] = True
        assert directives["test"]

    def test_del_directive(self, directives):
        directives["test"] = True
        directives._set_defined_directive(_NP._name, 100)
        assert directives[_NP._name] == 100
        assert directives["test"]
        del directives[_NP._name]
        assert directives[_NP._name] == _NP._default
        del directives["test"]
        with pytest.raises(KeyError):
            directives["test"]

    def test_update_directive_without_aggregate(
        self, directives, non_default_directive_values
    ):
        valid_values_1 = non_default_directive_values[1]
        expected_values = {
            "np": 4,
            "ngpu": 1,
            "nranks": 0,
            "omp_num_threads": 10,
            "executable": "PathFinder",
            "walltime": datetime.timedelta(hours=20.0),
            "memory": 16,
            "processor_fraction": 0.5,
        }
        directives.update(valid_values_1)
        for dirs in directives:
            assert directives[dirs] == expected_values[dirs]

    def test_update_directive_serial(
        self, available_directives_list, non_default_directive_values
    ):
        directives1 = _Directives(available_directives_list)
        directives2 = _Directives(available_directives_list)
        valid_values_0 = non_default_directive_values[0]
        valid_values_1 = non_default_directive_values[1]
        expected_values = {
            "np": 100,
            "ngpu": 10,
            "nranks": 5,
            "omp_num_threads": 20,
            "executable": "Non Default Path",
            "walltime": datetime.timedelta(hours=84.0),
            "memory": 32,
            "processor_fraction": 0.5,
        }
        directives1.update(valid_values_0)
        directives2.update(valid_values_1)
        directives1.update(directives2, aggregate=True)
        for dirs in directives1:
            assert directives1[dirs] == expected_values[dirs]

    def test_update_directive_parallel(
        self, available_directives_list, non_default_directive_values
    ):
        directives1 = _Directives(available_directives_list)
        directives2 = _Directives(available_directives_list)
        valid_values_0 = non_default_directive_values[0]
        valid_values_1 = non_default_directive_values[1]
        expected_values = {
            "np": 104,
            "ngpu": 11,
            "nranks": 5,
            "omp_num_threads": 30,
            "executable": "Non Default Path",
            "walltime": datetime.timedelta(hours=64.0),
            "memory": 48,
            "processor_fraction": 0.5,
        }
        directives1.update(valid_values_0)
        directives2.update(valid_values_1)
        directives1.update(directives2, aggregate=True, parallel=True)
        for dirs in directives1:
            assert directives1[dirs] == expected_values[dirs]

    def test_evaluate_directive_none_job(
        self, directives, non_default_directive_values
    ):
        directives.evaluate(None)
        valid_values = non_default_directive_values[0]
        valid_values["processor_fraction"] = lambda job: job.sp.i / 10
        directives.update(valid_values)
        with pytest.raises(RuntimeError):
            directives.evaluate(None)

    def test_evaluate_directive_valid_job(
        self, available_directives_list, non_default_directive_values
    ):
        _tmp_dir = TemporaryDirectory(prefix="flow-directives_")
        FlowProject.init_project(path=_tmp_dir.name)
        project = FlowProject.get_project(path=_tmp_dir.name)
        for i in range(5):
            project.open_job(dict(i=i)).init()

        valid_values = non_default_directive_values[0]
        valid_values["processor_fraction"] = lambda job: round(job.sp.i / 10, 1)

        for job in project:
            directives = _Directives(available_directives_list)
            directives.update(
                {"processor_fraction": lambda job: round(job.sp.i / 10, 1)}
            )
            directives.evaluate((job,))
            assert directives["processor_fraction"] == round(job.sp.i / 10, 1)



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Environments for the University of Michigan HPC environment."""
from ..environment import DefaultSlurmEnvironment


class GreatLakesEnvironment(DefaultSlurmEnvironment):
    """Environment profile for the Great Lakes supercomputer.

    https://arc-ts.umich.edu/greatlakes/
    """

    hostname_pattern = r"gl(-login)?[0-9]+\.arc-ts\.umich\.edu"
    template = "umich-greatlakes.sh"
    cores_per_node = 1

    @classmethod
    def add_args(cls, parser):
        """Add arguments to parser.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser where arguments will be added.

        """
        super().add_args(parser)
        parser.add_argument(
            "--partition",
            choices=("standard", "gpu", "largemem"),
            default="standard",
            help="Specify the partition to submit to. (default=standard)",
        )


__all__ = ["GreatLakesEnvironment"]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Contains logic for working with flow related information stored in the signac config."""

from ..errors import ConfigKeyError

_FLOW_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {
        "status_performance_warn_threshold": {"type": "number"},
        "eligible_jobs_max_lines": {"type": "integer"},
        "status_parallelization": {"type": "string"},
    },
}

_FLOW_CONFIG_DEFAULTS = {
    "status_performance_warn_threshold": 0.2,
    "eligible_jobs_max_lines": 10,
    "status_parallelization": "none",
}


class _GetConfigValueNoneType:
    pass


_GET_CONFIG_VALUE_NONE = _GetConfigValueNoneType()


def get_config_value(project, key, ns=None, default=_GET_CONFIG_VALUE_NONE):
    """Request a value from the user's configuration, failing if not available.

    Parameters
    ----------
    project : flow.FlowProject
        The project whose config is queried.
    key : str
        The environment specific configuration key.
    ns : str
        The namespace in which to look for the key. (Default value = None)
    default
        A default value in case the key cannot be found within the user's configuration. If not
        provided, the behavior if the key is not present is to raise a :class:`~.ConfigKeyError`
        exception.

    Returns
    -------
    object
        The value or default value.

    Raises
    ------
    :class:`~.ConfigKeyError`
        If the key is not in the user's configuration and no default value is provided.

    """
    try:
        if ns is None:
            return project.config["flow"][key]
        else:
            return project.config["flow"][ns][key]
    except KeyError:
        if default is _GET_CONFIG_VALUE_NONE:
            k = str(key) if ns is None else f"{ns}.{key}"
            raise ConfigKeyError("flow." + k)
        else:
            return default



# Copyright (c) 2021 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import os
import subprocess
import sys
from tempfile import TemporaryDirectory

import pytest
import signac

import flow


class ExitCodeError(RuntimeError):
    pass


class TestCLI:
    @pytest.fixture(autouse=True)
    def setUp(self, request):
        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is None:
            pythonpath = [os.getcwd()]
        else:
            pythonpath = [os.getcwd()] + pythonpath.split(":")
        os.environ["PYTHONPATH"] = ":".join(pythonpath)
        self.tmpdir = TemporaryDirectory(prefix="signac-flow_")
        request.addfinalizer(self.tmpdir.cleanup)
        self.cwd = os.getcwd()
        os.chdir(self.tmpdir.name)
        request.addfinalizer(self.return_to_cwd)

    def return_to_cwd(self):
        os.chdir(self.cwd)

    def call(self, command, input=None, shell=False, error=False, raise_error=True):
        p = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=shell,
        )
        if input:
            p.stdin.write(input.encode())
        out, err = p.communicate()
        if p.returncode != 0 and raise_error:
            raise ExitCodeError(f"STDOUT='{out}' STDERR='{err}'")
        return err.decode() if error else out.decode()

    def test_print_usage(self):
        with pytest.raises(ExitCodeError):
            self.call("python -m flow".split())

        out = self.call("python -m flow".split(), raise_error=False)
        assert "usage:" in out

    def test_version(self):
        out = self.call("python -m flow --version".split())
        assert f"signac-flow {flow.__version__}" in out

    def test_help(self):
        out = self.call("python -m flow --help".split())
        assert "positional arguments:" in out
        # Help string changed in 3.10.
        py310_or_greater = sys.version_info >= (3, 10)
        assert ("options:" if py310_or_greater else "optional arguments:") in out

    def test_init_flowproject(self):
        self.call("python -m flow init".split())
        assert os.path.exists("project.py")

    def test_init_flowproject_alias(self):
        self.call("python -m flow init".split())
        assert os.path.exists("project.py")

    def test_init_flowproject_fail_if_exists(self):
        self.call("python -m flow init".split())
        assert os.path.exists("project.py")
        with pytest.raises(ExitCodeError):
            self.call("python -m flow init".split())

    @pytest.mark.parametrize("template", flow.template.TEMPLATES)
    def test_init_flowproject_template(self, template):
        self.call(f"python -m flow init -t {template}".split())
        assert os.path.exists("project.py")

    @pytest.mark.parametrize(
        "environment, template",
        [
            ("DefaultSlurmEnvironment", "slurm.sh"),
            ("DefaultPBSEnvironment", "pbs.sh"),
            ("Bridges2Environment", "bridges2.sh"),
            ("SummitEnvironment", "summit.sh"),
        ],
    )
    def test_template_create_base(self, environment, template):
        self.call("python -m flow init".split())
        # Explicitly set the environment via the SIGNAC_FLOW_ENVIRONMENT environment
        # variable for test consistency
        output = self.call(
            f"SIGNAC_FLOW_ENVIRONMENT='{environment}' python -m flow template create",
            error=True,
            shell=True,
        )
        assert os.path.exists("templates/script.sh")
        with open("templates/script.sh") as fh:
            custom_script_lines = fh.read().splitlines()
        assert f'{{% extends "{template}" %}}' == custom_script_lines[0]
        script_location = signac.get_project().fn("templates/script.sh")
        assert output == (
            f"Created user script template at '{script_location}' "
            f"extending the template '{template}'.\n"
        )

    def test_template_create_fail_on_recall(self):
        self.call("python -m flow init".split())
        self.call("python -m flow template create".split())
        assert os.path.exists("templates/script.sh")
        with pytest.raises(ExitCodeError):
            self.call("python -m flow template create".split())

    @pytest.mark.parametrize("name_arg", ("-n foo.sh", "--name foo.sh"))
    def test_template_create_custom_name(self, name_arg):
        self.call("python -m flow init".split())
        self.call(f"python -m flow template create {name_arg}".split())
        assert os.path.exists("templates/foo.sh")

    @pytest.mark.parametrize("extends_arg", ("-e slurm.sh", "--extends slurm.sh"))
    def test_template_create_custom_extends(self, extends_arg):
        self.call("python -m flow init".split())
        self.call(f"python -m flow template create {extends_arg}".split())
        assert os.path.exists("templates/script.sh")
        with open("templates/script.sh") as fh:
            custom_script_lines = fh.read().splitlines()
        assert '{% extends "slurm.sh" %}' in custom_script_lines[0]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import io
import operator
import os
import sys

import generate_template_reference_data as gen
import pytest
import signac
from test_project import redirect_stderr, redirect_stdout

import flow
import flow.environments
from flow.scheduling.fake_scheduler import FakeScheduler


def _env_name(env):
    name = f"{env.__module__}.{env.__name__}"
    return ".".join(name.split(".")[1:])


def find_envs():
    """Yields the environments to be tested."""
    for name, env in flow.environment.ComputeEnvironment.registry.items():
        if env.__module__.startswith("flow.environments"):
            yield env


@pytest.mark.parametrize("env", find_envs())
def test_env(env, monkeypatch):
    monkeypatch.setattr(flow.FlowProject, "_store_bundled", gen._store_bundled)
    # We need to set the scheduler manually. The FakeScheduler is used for two
    # reasons. First, the FakeScheduler prints scripts to screen on submission
    # and we can capture that output. Second, the FakeScheduler won't try to
    # call any cluster executable (e.g. squeue) associated with the real
    # schedulers used on supported clusters. Otherwise submission would fail
    # when attempting to determine what jobs already exist on the scheduler.
    monkeypatch.setattr(env, "scheduler_type", FakeScheduler)

    # Force asserts to show the full file when failures occur.
    # Useful to debug errors that arise.

    # Must import the data into the project.
    with signac.TemporaryProject() as p:
        with gen.get_masked_flowproject(p, environment=env) as fp:
            # Here we set the appropriate executable for all the operations. This
            # is necessary as otherwise the default executable between submitting
            # and running could look different depending on the environment.
            for group in fp.groups.values():
                for op_key in group.operations:
                    if op_key in group.operation_directives:
                        monkeypatch.setitem(
                            group.operation_directives[op_key],
                            "executable",
                            gen.MOCK_EXECUTABLE,
                        )
            fp.import_from(origin=gen.ARCHIVE_DIR)
            jobs = fp.find_jobs(dict(environment=_env_name(env)))
            if not len(jobs):
                raise RuntimeError(
                    f"No reference data for environment {_env_name(env)}!"
                )
            reference = []
            generated = []
            for job in jobs:
                parameters = job.sp.parameters()
                if "bundle" in parameters:
                    bundle = parameters.pop("bundle")
                    tmp_out = io.TextIOWrapper(io.BytesIO(), sys.stdout.encoding)
                    with open(os.devnull, "w") as devnull:
                        with redirect_stderr(devnull):
                            with redirect_stdout(tmp_out):
                                fp.submit(
                                    jobs=[job],
                                    names=bundle,
                                    pretend=True,
                                    force=True,
                                    bundle_size=len(bundle),
                                    **parameters,
                                )
                    tmp_out.seek(0)
                    msg = f"---------- Bundled submission of job {job}"
                    generated.extend([msg] + tmp_out.read().splitlines())

                    with open(job.fn("script_{}.sh".format("_".join(bundle)))) as file:
                        reference.extend([msg] + file.read().splitlines())
                else:
                    for op in {**fp.operations, **fp.groups}:
                        if "partition" in parameters:
                            # Don't try to submit GPU operations to CPU partitions
                            # and vice versa.  We should be able to relax this
                            # requirement if we make our error checking more
                            # consistent.
                            if operator.xor(
                                "gpu" in parameters["partition"].lower(),
                                "gpu" in op.lower(),
                            ):
                                continue
                        tmp_out = io.TextIOWrapper(io.BytesIO(), sys.stdout.encoding)
                        with open(os.devnull, "w") as devnull:
                            with redirect_stderr(devnull):
                                with redirect_stdout(tmp_out):
                                    fp.submit(
                                        jobs=[job],
                                        names=[op],
                                        pretend=True,
                                        force=True,
                                        **parameters,
                                    )
                        tmp_out.seek(0)
                        msg = f"---------- Submission of operation {op} for job {job}."
                        generated.extend([msg] + tmp_out.read().splitlines())

                        with open(job.fn(f"script_{op}.sh")) as file:
                            reference.extend([msg] + file.read().splitlines())
            assert "\n".join(generated) == "\n".join(reference)



import pytest
from conftest import TestProjectBase

from flow import FlowProject
from flow.errors import ConfigKeyError
from flow.util.config import get_config_value


class _ConfigProject(FlowProject):
    pass


class TestConfigFilter(TestProjectBase):
    project_class = _ConfigProject

    def test_base(self):
        project = self.mock_project(config_overrides={"flow": {"foo": "test_string"}})
        assert get_config_value(project, "foo") == "test_string"

    def test_missing_key(self):
        project = self.mock_project()
        with pytest.raises(ConfigKeyError):
            get_config_value(project, "foo")

    def test_missing_key_with_default(self):
        project = self.mock_project()
        default = "default_string"
        assert get_config_value(project, "foo", default=default) == default

    def test_set_key_with_default(self):
        default = "default_string"
        set_value = "test_string"
        project = self.mock_project(config_overrides={"flow": {"foo": set_value}})
        assert get_config_value(project, "foo", default=default) == set_value

    def test_with_namespace(self):
        project = self.mock_project(
            config_overrides={"flow": {"bar": {"foo": "test_string"}}}
        )
        assert get_config_value(project, "foo", ns="bar") == "test_string"

    def test_missing_key_with_default_with_namespace(self):
        default = "default_string"
        project = self.mock_project()
        assert get_config_value(project, "foo", ns="bar", default=default) == default

    def test_set_key_with_default_with_namespace(self):
        default = "default_string"
        set_value = "test_string"
        project = self.mock_project(
            config_overrides={"flow": {"bar": {"foo": set_value}}}
        )
        assert get_config_value(project, "foo", ns="bar", default=default) == set_value



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.

from flow import get_environment
from flow.environment import ComputeEnvironment, TestEnvironment


class TestProject:
    def test_get_TestEnvironment(self):
        env = get_environment()
        assert issubclass(env, ComputeEnvironment)
        assert not issubclass(env, TestEnvironment)
        env = get_environment(test=True)
        assert issubclass(env, TestEnvironment)



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Environments for INCITE supercomputers.

http://www.doeleadershipcomputing.org/
"""
from math import ceil, gcd

from ..environment import (
    DefaultLSFEnvironment,
    DefaultSlurmEnvironment,
    template_filter,
)
from ..util.template_filters import check_utilization


class SummitEnvironment(DefaultLSFEnvironment):
    """Environment profile for the Summit supercomputer.

    Example::

        @Project.operation(directives={
            "nranks": 3,  # 3 MPI ranks per operation
            "ngpu": 3,  # 3 GPUs
            "np": 3,  # 3 CPU cores
            "rs_tasks": 3,  # 3 tasks per resource set
            "extra_jsrun_args": '--smpiargs="-gpu"',  # extra jsrun arguments
        })
        def my_operation(job):
            ...

    https://www.olcf.ornl.gov/summit/
    """

    hostname_pattern = r".*\.summit\.olcf\.ornl\.gov"
    template = "summit.sh"
    mpi_cmd = "jsrun"
    cores_per_node = 42
    gpus_per_node = 6

    @template_filter
    def calc_num_nodes(cls, resource_sets, parallel=False):
        """Compute the number of nodes needed.

        Parameters
        ----------
        resource_sets : iterable of tuples
            Resource sets for each operation, as a sequence of tuples of
            *(Number of resource sets, tasks (MPI Ranks) per resource set,
            physical cores (CPUs) per resource set, GPUs per resource set)*.
        parallel : bool
            Whether operations should run in parallel or serial. (Default value
            = False)

        Returns
        -------
        int
            Number of nodes needed.

        """
        nodes_used_final = 0
        cores_used = gpus_used = nodes_used = 0
        for nsets, tasks, cpus_per_task, gpus in resource_sets:
            if not parallel:
                # In serial mode we reset for every operation.
                cores_used = gpus_used = nodes_used = 0
            for _ in range(nsets):
                cores_used += tasks * cpus_per_task
                gpus_used += gpus
                while cores_used > cls.cores_per_node or gpus_used > cls.gpus_per_node:
                    nodes_used += 1
                    cores_used = max(0, cores_used - cls.cores_per_node)
                    gpus_used = max(0, gpus_used - cls.gpus_per_node)
            if not parallel:
                #  Note that when running in serial the "leftovers" must be
                #  accounted for on a per-operation basis.
                if cores_used > 0 or gpus_used > 0:
                    nodes_used += 1
                nodes_used_final = max(nodes_used, nodes_used_final)
        if parallel:
            if cores_used > 0 or gpus_used > 0:
                nodes_used += 1
            nodes_used_final = nodes_used
        return nodes_used_final

    @template_filter
    def guess_resource_sets(cls, operation):
        """Determine the resources sets needed for an operation.

        Parameters
        ----------
        operation : :class:`flow.BaseFlowOperation`
            The operation whose directives will be used to compute the resource
            set.

        Returns
        -------
        int
            Number of resource sets.
        int
            Number of tasks (MPI ranks) per resource set.
        int
            Number of physical cores (CPUs) per resource set.
        int
            Number of GPUs per resource set.

        """
        ntasks = max(operation.directives.get("nranks", 1), 1)
        np = operation.directives.get("np", ntasks)

        cpus_per_task = max(operation.directives.get("omp_num_threads", 1), 1)
        # separate OMP threads (per resource sets) from tasks
        np //= cpus_per_task

        np_per_task = max(1, np // ntasks)
        ngpu = operation.directives.get("ngpu", 0)
        g = gcd(ngpu, ntasks)
        if ngpu >= ntasks:
            nsets = ngpu // (ngpu // g)
        else:
            nsets = ntasks // (ntasks // g)

        tasks_per_set = max(ntasks // nsets, 1)
        tasks_per_set = max(tasks_per_set, operation.directives.get("rs_tasks", 1))

        gpus_per_set = ngpu // nsets
        cpus_per_set = tasks_per_set * cpus_per_task * np_per_task

        return nsets, tasks_per_set, cpus_per_set, gpus_per_set

    @classmethod
    def jsrun_options(cls, resource_set):
        """Return jsrun options for the provided resource set.

        Parameters
        ----------
        resource_set : tuple
            Tuple of *(Number of resource sets, tasks (MPI Ranks) per resource
            set, physical cores (CPUs) per resource set, GPUs per resource
            set)*.

        Returns
        -------
        str
            Resource set options.

        """
        nsets, tasks, cpus, gpus = resource_set
        cuda_aware_mpi = (
            "--smpiargs='-gpu'" if (nsets > 0 or tasks > 0) and gpus > 0 else ""
        )
        return f"-n {nsets} -a {tasks} -c {cpus} -g {gpus} {cuda_aware_mpi}"

    @classmethod
    def _get_mpi_prefix(cls, operation, parallel):
        """Get the jsrun options based on directives.

        Parameters
        ----------
        operation : :class:`flow.project._JobOperation`
            The operation to be prefixed.
        parallel : bool
            If True, operations are assumed to be executed in parallel, which
            means that the number of total tasks is the sum of all tasks
            instead of the maximum number of tasks. Default is set to False.

        Returns
        -------
        str
            The prefix to be added to the operation's command.

        """
        extra_args = str(operation.directives.get("extra_jsrun_args", ""))
        resource_set = cls.guess_resource_sets(operation)
        mpi_prefix = cls.mpi_cmd + " " + cls.jsrun_options(resource_set)
        mpi_prefix += " -d packed -b rs " + extra_args + (" " if extra_args else "")
        return mpi_prefix


class AndesEnvironment(DefaultSlurmEnvironment):
    """Environment profile for the Andes supercomputer.

    https://www.olcf.ornl.gov/olcf-resources/compute-systems/andes/
    """

    hostname_pattern = r"andes-.*\.olcf\.ornl\.gov"
    template = "andes.sh"
    mpi_cmd = "srun"
    cores_per_node = 32

    @classmethod
    def add_args(cls, parser):
        """Add arguments to parser.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser where arguments will be added.

        """
        super().add_args(parser)
        parser.add_argument(
            "--partition",
            choices=["batch", "gpu"],
            default="batch",
            help="Specify the partition to submit to.",
        )


class CrusherEnvironment(DefaultSlurmEnvironment):
    """Environment profile for the Cluster supercomputer.

    https://docs.olcf.ornl.gov/systems/crusher_quick_start_guide.html
    """

    hostname_pattern = r".*\.crusher\.olcf\.ornl\.gov"
    template = "crusher.sh"
    cores_per_node = 56
    gpus_per_node = 8
    mpi_cmd = "srun"

    @template_filter
    def calc_num_nodes(cls, ngpus, ncpus, threshold):
        """Compute the number of nodes needed to meet the resource request.

        Also raise an error when the requested resource do not come close to saturating the asked
        for nodes.
        """
        nodes_gpu = max(1, int(ceil(ngpus / cls.gpus_per_node)))
        nodes_cpu = max(1, int(ceil(ncpus / cls.cores_per_node)))
        if nodes_gpu >= nodes_cpu:
            check_utilization(nodes_gpu, ngpus, cls.gpus_per_node, threshold, "compute")
            return nodes_gpu
        check_utilization(nodes_cpu, ncpus, cls.cores_per_node, threshold, "compute")
        return nodes_cpu

    @classmethod
    def _get_mpi_prefix(cls, operation, parallel):
        """Get the correct srun command for the job.

        We don't currently support CPU/GPU mapping and expect the program to do this in code.
        """
        nranks = operation.directives.get("nranks", 0)
        if nranks == 0:
            return ""
        ngpus = operation.directives["ngpu"]
        np = operation.directives.get("np", 1)
        omp_num_threads = max(operation.directives.get("omp_num_threads", 1), 1)
        mpi_np_calc = nranks * omp_num_threads
        if np > 1 and nranks > 1 and np != mpi_np_calc:
            raise RuntimeWarning(
                f"Using provided value for np={np}, which seems incompatible with MPI directives "
                f"{mpi_np_calc}."
            )
        base_str = f"{cls.mpi_cmd} --ntasks={nranks}"
        threads = max(omp_num_threads, np) if nranks == 1 else max(1, omp_num_threads)
        base_str += f" --cpus-per-task={threads} --gpus={ngpus}"
        return base_str


__all__ = ["SummitEnvironment", "AndesEnvironment", "CrusherEnvironment"]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""The main command line interface for signac-flow.

The interface is accessible via the `flow` command and allows users to
initialize FlowProject class definitions directly from the command line.

Execute `flow --help` for more information.
"""
import argparse
import errno
import logging
import os
import string
import sys

import jinja2
from signac import get_project, init_project

from . import __version__, environment, template

logger = logging.getLogger(__name__)


def main_init(args):
    """Initialize a FlowProject from a template.

    The available templates are defined in the template module.
    """
    try:
        get_project()
    except LookupError:
        init_project()
        print("Initialized signac project in current directory.", file=sys.stderr)
    try:
        return template.init(template=args.template)
    except OSError as error:
        raise RuntimeError(
            f"Error occurred while trying to initialize a flow project: {error}"
        )


def main_template_create(args):
    """Create custom template based on a specified or detected environment."""
    if args.extends is None:
        extend_template = environment.get_environment().template
    else:
        extend_template = args.extends[0]

    project = get_project()
    template_directory = project.fn(project._config.get("template_dir", "templates"))
    os.makedirs(template_directory, exist_ok=True)

    # grab and render custom template
    jinja_env = jinja2.Environment(loader=jinja2.PackageLoader("flow"))
    custom_template = jinja_env.get_template("custom_script_template.sh")
    new_template = custom_template.render(extend_template=extend_template).strip(
        string.whitespace
    )

    # If name is the default then it is a string if not it gets passed as a list
    script_name = args.name if isinstance(args.name, str) else args.name[0]
    script_path = os.path.join(template_directory, script_name)
    try:
        with open(script_path, "x") as fh:
            fh.write(new_template)
    except OSError as error:
        if error.errno == errno.EEXIST:
            logger.error(
                f"Error while trying to create custom template. Delete '{script_path}' "
                f"first and rerun command."
            )
        else:
            logger.error(f"Error while trying to create custom template: '{error}'.")
        raise
    else:
        print(
            f"Created user script template at '{script_path}' extending the template "
            f"'{extend_template}'.",
            file=sys.stderr,
        )


def main():
    """Define the 'flow' command line interface.

    This function defines the main 'flow' command line interface which can be used
    to initialize FlowProject modules from different templates as well as print the
    version of the installed signac-flow package.
    """
    parser = argparse.ArgumentParser(
        description="flow provides the basic components to set up workflows for "
        "projects as part of the signac framework."
    )
    parser.add_argument(
        "--debug", action="store_true", help="Show traceback on error for debugging."
    )
    parser.add_argument(
        "--version", action="store_true", help="Display the version number and exit."
    )

    subparsers = parser.add_subparsers()

    # the flow init command
    parser_init = subparsers.add_parser(
        "init", help="Initialize a signac-flow project."
    )
    parser_init.set_defaults(func=main_init)
    parser_init.add_argument(
        "-t",
        "--template",
        type=str,
        choices=tuple(sorted(template.TEMPLATES)),
        default="minimal",
        help="Specify a template to use. Default value: 'minimal'.",
    )

    # the flow template command
    parser_template = subparsers.add_parser(
        "template", help="Create and manage custom user templates."
    )
    template_subparsers = parser_template.add_subparsers()

    # flow template create command
    parser_template_create = template_subparsers.add_parser(
        "create",
        help="Create a new custom template based on the detected or selected environment.",
    )
    parser_template_create.add_argument(
        "--extends",
        "-e",
        nargs=1,
        type=str,
        default=None,
        help="Optionally specify a template to extend (including the file extension if present). "
        "If not provided, the current environment's template is used. For systems "
        "without a scheduler (or an unsupported scheduler) this is 'base_script.sh'.",
    )

    parser_template_create.add_argument(
        "--name",
        "-n",
        nargs=1,
        type=str,
        default="script.sh",
        help="Optionally specify a name to use as the the name of the custom user "
        "template. Defaults to 'script.sh', which will be used by default in flow "
        "submissions for the current project.",
    )

    parser_template_create.set_defaults(func=main_template_create)

    if "--version" in sys.argv:
        print("signac-flow", __version__)
        sys.exit(0)

    args = parser.parse_args()
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    if not hasattr(args, "func"):
        parser.print_usage()
        sys.exit(2)
    try:
        args.func(args)
    except KeyboardInterrupt:
        sys.stderr.write("\n")
        sys.stderr.write("Interrupted.\n")
        if args.debug:
            raise
        sys.exit(1)
    except Exception as error:
        sys.stderr.write(f"{str(error)}\n")
        if args.debug:
            raise
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Operation hooks."""
from .hooks import _Hooks

__all__ = ["_Hooks"]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Defines the API for the util subpackage."""
from . import misc, mistune, template_filters, translate

__all__ = [
    "misc",
    "mistune",
    "template_filters",
    "translate",
]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""FlowProject module templates.

These templates can be initialized via the init() function defined
in this module and the main 'flow' command line interface.
"""
import errno
import logging
import os
import sys

import jinja2

logger = logging.getLogger(__name__)


TEMPLATES = {
    "minimal": [
        ("project.py", "project_minimal.pyt"),
    ],
    "example": [
        ("project.py", "project_example.pyt"),
    ],
    "testing": [
        ("project.py", "project_testing.pyt"),
    ],
}


def init(alias=None, template=None, path=None):
    """Initialize a templated :class:`~.FlowProject` module.

    Parameters
    ----------
    alias : str
        Python identifier used as a file name for the template output. Uses
        ``"project"`` if None.  (Default value = None)

    template : str
        Name of the template to use. Built-in templates are:

        * ``"minimal"``
        * ``"example"``
        * ``"testing"``

        Uses ``"minimal"`` if None. (Default value = None)

    path : str
        Directory where the output file is placed. Uses the current working
        directory if None. (Default value = None)

    Returns
    -------
    list
        List of file names created.

    """
    if alias is None:
        alias = "project"
    elif not alias.isidentifier():
        raise ValueError(
            f"The alias '{alias}' is not a valid Python identifier and therefore "
            "cannot be used as a FlowProject alias."
        )
    if template is None:
        template = "minimal"

    if os.path.splitext(alias)[1]:
        raise RuntimeError("Please provide a name without suffix!")

    project_class_name = alias.capitalize()
    if not project_class_name.endswith("Project"):
        project_class_name += "Project"

    template_environment = jinja2.Environment(
        loader=jinja2.ChoiceLoader(
            [
                jinja2.FileSystemLoader("templates"),
                jinja2.PackageLoader("flow", "templates"),
            ]
        ),
        trim_blocks=True,
        lstrip_blocks=True,
    )

    context = {}
    context["alias"] = alias
    context["project_class_name"] = project_class_name

    # render all templates
    codes = {}

    for fn, fn_template in TEMPLATES[template]:
        fn_ = fn.format(alias=alias)  # some of the filenames may depend on the alias
        template = template_environment.get_template(fn_template)
        codes[fn_] = template.render(**context)

    # create files
    files_created = []
    for fn, code in codes.items():
        try:
            if path is not None:
                fn = os.path.join(path, fn)
            with open(fn, "x") as fw:
                fw.write(code + "\n")
        except OSError as error:
            error_message = (
                f"Error while creating FlowProject module with name '{alias}': "
            )
            if error.errno == errno.EEXIST:
                error_message += f"a file named '{fn}' already exists!"
            else:
                error_message += f"'{error}'."
            raise OSError(error_message)
        else:
            files_created.append(fn)
            print(f"Created file '{fn}'.", file=sys.stderr)
    return files_created



import pytest

from flow.util.misc import _bidict
from flow.util.template_filters import calc_memory


class TestBidict:
    @pytest.fixture()
    def bd(self):
        return _bidict({"a": 1, "b": 2})

    def test_basic_dict(self, bd):
        bd["c"] = 1
        assert bd == {"a": 1, "c": 1, "b": 2}
        assert bd.inverse == {1: ["a", "c"], 2: ["b"]}
        assert len(bd) == 3
        assert list(bd.keys()) == ["a", "b", "c"]
        assert list(bd.values()) == [1, 2, 1]
        assert list(bd.items()) == [("a", 1), ("b", 2), ("c", 1)]

    def test_get_func(self, bd):
        assert bd.get("a") == 1
        assert bd.get("q", 42) == 42

    def test_contains(self, bd):
        assert "a" in bd
        assert "q" not in bd

    def test_delitem(self, bd):
        bd["c"] = 1
        del bd["c"]
        assert bd == {"a": 1, "b": 2}
        assert bd.inverse == {1: ["a"], 2: ["b"]}
        assert len(bd) == 2

    def test_setitem(self, bd):
        bd["b"] = 3
        assert bd == {"a": 1, "b": 3}
        assert bd.inverse == {3: ["b"], 1: ["a"]}

    def test_update(self, bd):
        bd.update({"c": 2})
        assert bd == {"b": 2, "c": 2, "a": 1}
        assert bd.inverse == {1: ["a"], 2: ["b", "c"]}

    def test_popitem(self, bd):
        # Note that MutableMapping's popitem is *not* LIFO-ordered, it pops
        # according to the key iteration order (which means it defaults to FIFO
        # for the underlying dict, which is insertion-ordered).
        assert bd.popitem() == ("a", 1)
        assert bd == {"b": 2}
        assert bd.inverse == {2: ["b"]}

    def test_setdefault_existing_key(self, bd):
        bd.setdefault("b", 7)
        assert bd == {"b": 2, "a": 1}
        assert bd.inverse == {1: ["a"], 2: ["b"]}

    def test_setdefault_nonexistant_key(self, bd):
        bd.setdefault("c", 9)
        assert bd == {"a": 1, "b": 2, "c": 9}
        assert bd.inverse == {2: ["b"], 9: ["c"], 1: ["a"]}

    def test_pop(self, bd):
        assert bd.pop("b") == 2
        assert bd == {"a": 1}
        assert bd.inverse == {1: ["a"]}

    def test_clear(self, bd):
        bd.clear()
        assert bd == {}
        assert bd.inverse == {}


class TestTemplateFilters:
    @pytest.fixture()
    def mock_operations(self):
        class MockOp:
            def __init__(self, memory=None):
                self.directives = {"memory": memory}

        return [MockOp(1), MockOp(8), MockOp()]

    def test_calc_memory_serial(self, mock_operations):
        # Test when operations run in serial
        assert calc_memory([mock_operations[0]], False) == 1
        assert (
            calc_memory(
                [mock_operations[0], mock_operations[1], mock_operations[2]], False
            )
            == 8
        )

    def test_calc_memory_parallel(self, mock_operations):
        # Test when operations run in parallel
        assert calc_memory([mock_operations[0], mock_operations[1]], True) == 9
        assert calc_memory([mock_operations[2]], True) == 0
        assert (
            calc_memory(
                [mock_operations[0], mock_operations[1], mock_operations[2]], True
            )
            == 9
        )



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implementation of the scheduling system for a fake scheduler.

The :class:`~.FakeScheduler` class can be used in place of a real scheduler
to test the cluster job submission workflow.
"""
import logging

from .base import Scheduler

logger = logging.getLogger(__name__)


class FakeScheduler(Scheduler):
    """Implementation of the abstract :class:`~.Scheduler` class for a fake scheduler.

    This scheduler does not actually schedule (or execute) any jobs, but it can be used
    to test the submission workflow.
    """

    def jobs(self):
        """Return None (no jobs are scheduled by the FakeScheduler)."""
        return None

    def submit(self, script, **kwargs):
        r"""Print the script to screen.

        Parameters
        ----------
        script : str
            Script to print.
        \*\*kwargs
            Keyword arguments (ignored).

        """
        print(script)

    @classmethod
    def is_present(cls):
        """Return False.

        The :class:`~.FakeScheduler` is never present unless manually specified.
        """
        return False



# -*- coding: utf-8 -*-
"""Copyright (c) 2011-2020 Sergey Astanin and contributors

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Pretty-print tabular data."""

from __future__ import print_function
from __future__ import unicode_literals
from collections import namedtuple
from platform import python_version_tuple
import re
import math


from collections.abc import Iterable
from itertools import zip_longest as izip_longest
from functools import reduce, partial

_none_type = type(None)
_bool_type = bool
_int_type = int
_long_type = int
_float_type = float
_text_type = str
_binary_type = bytes
basestring = str

import io

def _is_file(f):
    return isinstance(f, io.IOBase)


try:
    import wcwidth  # optional wide-character (CJK) support
except ImportError:
    wcwidth = None

try:
    from html import escape as htmlescape
except ImportError:
    from cgi import escape as htmlescape


__all__ = ["tabulate", "tabulate_formats", "simple_separated_format"]
__version__ = "0.8.7"


# minimum extra space in headers
MIN_PADDING = 2

# Whether or not to preserve leading/trailing whitespace in data.
PRESERVE_WHITESPACE = False

_DEFAULT_FLOATFMT = "g"
_DEFAULT_MISSINGVAL = ""


# if True, enable wide-character (CJK) support
WIDE_CHARS_MODE = wcwidth is not None


Line = namedtuple("Line", ["begin", "hline", "sep", "end"])


DataRow = namedtuple("DataRow", ["begin", "sep", "end"])


# A table structure is suppposed to be:
#
#     --- lineabove ---------
#         headerrow
#     --- linebelowheader ---
#         datarow
#     --- linebewteenrows ---
#     ... (more datarows) ...
#     --- linebewteenrows ---
#         last datarow
#     --- linebelow ---------
#
# TableFormat's line* elements can be
#
#   - either None, if the element is not used,
#   - or a Line tuple,
#   - or a function: [col_widths], [col_alignments] -> string.
#
# TableFormat's *row elements can be
#
#   - either None, if the element is not used,
#   - or a DataRow tuple,
#   - or a function: [cell_values], [col_widths], [col_alignments] -> string.
#
# padding (an integer) is the amount of white space around data values.
#
# with_header_hide:
#
#   - either None, to display all table elements unconditionally,
#   - or a list of elements not to be displayed if the table has column headers.
#
TableFormat = namedtuple(
    "TableFormat",
    [
        "lineabove",
        "linebelowheader",
        "linebetweenrows",
        "linebelow",
        "headerrow",
        "datarow",
        "padding",
        "with_header_hide",
    ],
)


def _pipe_segment_with_colons(align, colwidth):
    """Return a segment of a horizontal line with optional colons which
    indicate column's alignment (as in `pipe` output format)."""
    w = colwidth
    if align in ["right", "decimal"]:
        return ("-" * (w - 1)) + ":"
    elif align == "center":
        return ":" + ("-" * (w - 2)) + ":"
    elif align == "left":
        return ":" + ("-" * (w - 1))
    else:
        return "-" * w


def _pipe_line_with_colons(colwidths, colaligns):
    """Return a horizontal line with optional colons to indicate column's
    alignment (as in `pipe` output format)."""
    if not colaligns:  # e.g. printing an empty data frame (github issue #15)
        colaligns = [""] * len(colwidths)
    segments = [_pipe_segment_with_colons(a, w) for a, w in zip(colaligns, colwidths)]
    return "|" + "|".join(segments) + "|"


def _mediawiki_row_with_attrs(separator, cell_values, colwidths, colaligns):
    alignment = {
        "left": "",
        "right": 'align="right"| ',
        "center": 'align="center"| ',
        "decimal": 'align="right"| ',
    }
    # hard-coded padding _around_ align attribute and value together
    # rather than padding parameter which affects only the value
    values_with_attrs = [
        " " + alignment.get(a, "") + c + " " for c, a in zip(cell_values, colaligns)
    ]
    colsep = separator * 2
    return (separator + colsep.join(values_with_attrs)).rstrip()


def _textile_row_with_attrs(cell_values, colwidths, colaligns):
    cell_values[0] += " "
    alignment = {"left": "<.", "right": ">.", "center": "=.", "decimal": ">."}
    values = (alignment.get(a, "") + v for a, v in zip(colaligns, cell_values))
    return "|" + "|".join(values) + "|"


def _html_begin_table_without_header(colwidths_ignore, colaligns_ignore):
    # this table header will be suppressed if there is a header row
    return "<table>\n<tbody>"


def _html_row_with_attrs(celltag, cell_values, colwidths, colaligns):
    alignment = {
        "left": "",
        "right": ' style="text-align: right;"',
        "center": ' style="text-align: center;"',
        "decimal": ' style="text-align: right;"',
    }
    values_with_attrs = [
        "<{0}{1}>{2}</{0}>".format(celltag, alignment.get(a, ""), htmlescape(c))
        for c, a in zip(cell_values, colaligns)
    ]
    rowhtml = "<tr>{}</tr>".format("".join(values_with_attrs).rstrip())
    if celltag == "th":  # it's a header row, create a new table header
        rowhtml = "<table>\n<thead>\n{}\n</thead>\n<tbody>".format(rowhtml)
    return rowhtml


def _moin_row_with_attrs(celltag, cell_values, colwidths, colaligns, header=""):
    alignment = {
        "left": "",
        "right": '<style="text-align: right;">',
        "center": '<style="text-align: center;">',
        "decimal": '<style="text-align: right;">',
    }
    values_with_attrs = [
        "{0}{1} {2} ".format(celltag, alignment.get(a, ""), header + c + header)
        for c, a in zip(cell_values, colaligns)
    ]
    return "".join(values_with_attrs) + "||"


def _latex_line_begin_tabular(colwidths, colaligns, booktabs=False):
    alignment = {"left": "l", "right": "r", "center": "c", "decimal": "r"}
    tabular_columns_fmt = "".join([alignment.get(a, "l") for a in colaligns])
    return "\n".join(
        [
            "\\begin{tabular}{" + tabular_columns_fmt + "}",
            "\\toprule" if booktabs else "\\hline",
        ]
    )


LATEX_ESCAPE_RULES = {
    r"&": r"\&",
    r"%": r"\%",
    r"$": r"\$",
    r"#": r"\#",
    r"_": r"\_",
    r"^": r"\^{}",
    r"{": r"\{",
    r"}": r"\}",
    r"~": r"\textasciitilde{}",
    "\\": r"\textbackslash{}",
    r"<": r"\ensuremath{<}",
    r">": r"\ensuremath{>}",
}


def _latex_row(cell_values, colwidths, colaligns, escrules=LATEX_ESCAPE_RULES):
    def escape_char(c):
        return escrules.get(c, c)

    escaped_values = ["".join(map(escape_char, cell)) for cell in cell_values]
    rowfmt = DataRow("", "&", "\\\\")
    return _build_simple_row(escaped_values, rowfmt)


def _rst_escape_first_column(rows, headers):
    def escape_empty(val):
        if isinstance(val, (_text_type, _binary_type)) and not val.strip():
            return ".."
        else:
            return val

    new_headers = list(headers)
    new_rows = []
    if headers:
        new_headers[0] = escape_empty(headers[0])
    for row in rows:
        new_row = list(row)
        if new_row:
            new_row[0] = escape_empty(row[0])
        new_rows.append(new_row)
    return new_rows, new_headers


_table_formats = {
    "simple": TableFormat(
        lineabove=Line("", "-", "  ", ""),
        linebelowheader=Line("", "-", "  ", ""),
        linebetweenrows=None,
        linebelow=Line("", "-", "  ", ""),
        headerrow=DataRow("", "  ", ""),
        datarow=DataRow("", "  ", ""),
        padding=0,
        with_header_hide=["lineabove", "linebelow"],
    ),
    "plain": TableFormat(
        lineabove=None,
        linebelowheader=None,
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("", "  ", ""),
        datarow=DataRow("", "  ", ""),
        padding=0,
        with_header_hide=None,
    ),
    "grid": TableFormat(
        lineabove=Line("+", "-", "+", "+"),
        linebelowheader=Line("+", "=", "+", "+"),
        linebetweenrows=Line("+", "-", "+", "+"),
        linebelow=Line("+", "-", "+", "+"),
        headerrow=DataRow("|", "|", "|"),
        datarow=DataRow("|", "|", "|"),
        padding=1,
        with_header_hide=None,
    ),
    "fancy_grid": TableFormat(
        lineabove=Line("╒", "═", "╤", "╕"),
        linebelowheader=Line("╞", "═", "╪", "╡"),
        linebetweenrows=Line("├", "─", "┼", "┤"),
        linebelow=Line("╘", "═", "╧", "╛"),
        headerrow=DataRow("│", "│", "│"),
        datarow=DataRow("│", "│", "│"),
        padding=1,
        with_header_hide=None,
    ),
    "github": TableFormat(
        lineabove=Line("|", "-", "|", "|"),
        linebelowheader=Line("|", "-", "|", "|"),
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("|", "|", "|"),
        datarow=DataRow("|", "|", "|"),
        padding=1,
        with_header_hide=["lineabove"],
    ),
    "pipe": TableFormat(
        lineabove=_pipe_line_with_colons,
        linebelowheader=_pipe_line_with_colons,
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("|", "|", "|"),
        datarow=DataRow("|", "|", "|"),
        padding=1,
        with_header_hide=["lineabove"],
    ),
    "orgtbl": TableFormat(
        lineabove=None,
        linebelowheader=Line("|", "-", "+", "|"),
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("|", "|", "|"),
        datarow=DataRow("|", "|", "|"),
        padding=1,
        with_header_hide=None,
    ),
    "jira": TableFormat(
        lineabove=None,
        linebelowheader=None,
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("||", "||", "||"),
        datarow=DataRow("|", "|", "|"),
        padding=1,
        with_header_hide=None,
    ),
    "presto": TableFormat(
        lineabove=None,
        linebelowheader=Line("", "-", "+", ""),
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("", "|", ""),
        datarow=DataRow("", "|", ""),
        padding=1,
        with_header_hide=None,
    ),
    "pretty": TableFormat(
        lineabove=Line("+", "-", "+", "+"),
        linebelowheader=Line("+", "-", "+", "+"),
        linebetweenrows=None,
        linebelow=Line("+", "-", "+", "+"),
        headerrow=DataRow("|", "|", "|"),
        datarow=DataRow("|", "|", "|"),
        padding=1,
        with_header_hide=None,
    ),
    "psql": TableFormat(
        lineabove=Line("+", "-", "+", "+"),
        linebelowheader=Line("|", "-", "+", "|"),
        linebetweenrows=None,
        linebelow=Line("+", "-", "+", "+"),
        headerrow=DataRow("|", "|", "|"),
        datarow=DataRow("|", "|", "|"),
        padding=1,
        with_header_hide=None,
    ),
    "rst": TableFormat(
        lineabove=Line("", "=", "  ", ""),
        linebelowheader=Line("", "=", "  ", ""),
        linebetweenrows=None,
        linebelow=Line("", "=", "  ", ""),
        headerrow=DataRow("", "  ", ""),
        datarow=DataRow("", "  ", ""),
        padding=0,
        with_header_hide=None,
    ),
    "mediawiki": TableFormat(
        lineabove=Line(
            '{| class="wikitable" style="text-align: left;"',
            "",
            "",
            "\n|+ <!-- caption -->\n|-",
        ),
        linebelowheader=Line("|-", "", "", ""),
        linebetweenrows=Line("|-", "", "", ""),
        linebelow=Line("|}", "", "", ""),
        headerrow=partial(_mediawiki_row_with_attrs, "!"),
        datarow=partial(_mediawiki_row_with_attrs, "|"),
        padding=0,
        with_header_hide=None,
    ),
    "moinmoin": TableFormat(
        lineabove=None,
        linebelowheader=None,
        linebetweenrows=None,
        linebelow=None,
        headerrow=partial(_moin_row_with_attrs, "||", header="'''"),
        datarow=partial(_moin_row_with_attrs, "||"),
        padding=1,
        with_header_hide=None,
    ),
    "youtrack": TableFormat(
        lineabove=None,
        linebelowheader=None,
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("|| ", " || ", " || "),
        datarow=DataRow("| ", " | ", " |"),
        padding=1,
        with_header_hide=None,
    ),
    "html": TableFormat(
        lineabove=_html_begin_table_without_header,
        linebelowheader="",
        linebetweenrows=None,
        linebelow=Line("</tbody>\n</table>", "", "", ""),
        headerrow=partial(_html_row_with_attrs, "th"),
        datarow=partial(_html_row_with_attrs, "td"),
        padding=0,
        with_header_hide=["lineabove"],
    ),
    "latex": TableFormat(
        lineabove=_latex_line_begin_tabular,
        linebelowheader=Line("\\hline", "", "", ""),
        linebetweenrows=None,
        linebelow=Line("\\hline\n\\end{tabular}", "", "", ""),
        headerrow=_latex_row,
        datarow=_latex_row,
        padding=1,
        with_header_hide=None,
    ),
    "latex_raw": TableFormat(
        lineabove=_latex_line_begin_tabular,
        linebelowheader=Line("\\hline", "", "", ""),
        linebetweenrows=None,
        linebelow=Line("\\hline\n\\end{tabular}", "", "", ""),
        headerrow=partial(_latex_row, escrules={}),
        datarow=partial(_latex_row, escrules={}),
        padding=1,
        with_header_hide=None,
    ),
    "latex_booktabs": TableFormat(
        lineabove=partial(_latex_line_begin_tabular, booktabs=True),
        linebelowheader=Line("\\midrule", "", "", ""),
        linebetweenrows=None,
        linebelow=Line("\\bottomrule\n\\end{tabular}", "", "", ""),
        headerrow=_latex_row,
        datarow=_latex_row,
        padding=1,
        with_header_hide=None,
    ),
    "tsv": TableFormat(
        lineabove=None,
        linebelowheader=None,
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("", "\t", ""),
        datarow=DataRow("", "\t", ""),
        padding=0,
        with_header_hide=None,
    ),
    "textile": TableFormat(
        lineabove=None,
        linebelowheader=None,
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("|_. ", "|_.", "|"),
        datarow=_textile_row_with_attrs,
        padding=1,
        with_header_hide=None,
    ),
}


tabulate_formats = list(sorted(_table_formats.keys()))

# The table formats for which multiline cells will be folded into subsequent
# table rows. The key is the original format specified at the API. The value is
# the format that will be used to represent the original format.
multiline_formats = {
    "plain": "plain",
    "simple": "simple",
    "grid": "grid",
    "fancy_grid": "fancy_grid",
    "pipe": "pipe",
    "orgtbl": "orgtbl",
    "jira": "jira",
    "presto": "presto",
    "pretty": "pretty",
    "psql": "psql",
    "rst": "rst",
}

# TODO: Add multiline support for the remaining table formats:
#       - mediawiki: Replace \n with <br>
#       - moinmoin: TBD
#       - youtrack: TBD
#       - html: Replace \n with <br>
#       - latex*: Use "makecell" package: In header, replace X\nY with
#         \thead{X\\Y} and in data row, replace X\nY with \makecell{X\\Y}
#       - tsv: TBD
#       - textile: Replace \n with <br/> (must be well-formed XML)

_multiline_codes = re.compile(r"\r|\n|\r\n")
_multiline_codes_bytes = re.compile(b"\r|\n|\r\n")
_invisible_codes = re.compile(
    r"\x1b\[\d+[;\d]*m|\x1b\[\d*\;\d*\;\d*m"
)  # ANSI color codes
_invisible_codes_bytes = re.compile(
    b"\x1b\\[\\d+\\[;\\d]*m|\x1b\\[\\d*;\\d*;\\d*m"
)  # ANSI color codes


def simple_separated_format(separator):
    """Construct a simple TableFormat with columns separated by a separator.

    >>> tsv = simple_separated_format("\\t") ; \
        tabulate([["foo", 1], ["spam", 23]], tablefmt=tsv) == 'foo \\t 1\\nspam\\t23'
    True

    """
    return TableFormat(
        None,
        None,
        None,
        None,
        headerrow=DataRow("", separator, ""),
        datarow=DataRow("", separator, ""),
        padding=0,
        with_header_hide=None,
    )


def _isconvertible(conv, string):
    try:
        conv(string)
        return True
    except (ValueError, TypeError):
        return False


def _isnumber(string):
    """
    >>> _isnumber("123.45")
    True
    >>> _isnumber("123")
    True
    >>> _isnumber("spam")
    False
    >>> _isnumber("123e45678")
    False
    >>> _isnumber("inf")
    True
    """
    if not _isconvertible(float, string):
        return False
    elif isinstance(string, (_text_type, _binary_type)) and (
        math.isinf(float(string)) or math.isnan(float(string))
    ):
        return string.lower() in ["inf", "-inf", "nan"]
    return True


def _isint(string, inttype=int):
    """
    >>> _isint("123")
    True
    >>> _isint("123.45")
    False
    """
    return (
        type(string) is inttype
        or (isinstance(string, _binary_type) or isinstance(string, _text_type))
        and _isconvertible(inttype, string)
    )


def _isbool(string):
    """
    >>> _isbool(True)
    True
    >>> _isbool("False")
    True
    >>> _isbool(1)
    False
    """
    return type(string) is _bool_type or (
        isinstance(string, (_binary_type, _text_type)) and string in ("True", "False")
    )


def _type(string, has_invisible=True, numparse=True):
    """The least generic type (type(None), int, float, str, unicode).

    >>> _type(None) is type(None)
    True
    >>> _type("foo") is type("")
    True
    >>> _type("1") is type(1)
    True
    >>> _type('\x1b[31m42\x1b[0m') is type(42)
    True
    >>> _type('\x1b[31m42\x1b[0m') is type(42)
    True

    """

    if has_invisible and (
        isinstance(string, _text_type) or isinstance(string, _binary_type)
    ):
        string = _strip_invisible(string)

    if string is None:
        return _none_type
    elif hasattr(string, "isoformat"):  # datetime.datetime, date, and time
        return _text_type
    elif _isbool(string):
        return _bool_type
    elif _isint(string) and numparse:
        return int
    elif _isint(string, _long_type) and numparse:
        return int
    elif _isnumber(string) and numparse:
        return float
    elif isinstance(string, _binary_type):
        return _binary_type
    else:
        return _text_type


def _afterpoint(string):
    """Symbols after a decimal point, -1 if the string lacks the decimal point.

    >>> _afterpoint("123.45")
    2
    >>> _afterpoint("1001")
    -1
    >>> _afterpoint("eggs")
    -1
    >>> _afterpoint("123e45")
    2

    """
    if _isnumber(string):
        if _isint(string):
            return -1
        else:
            pos = string.rfind(".")
            pos = string.lower().rfind("e") if pos < 0 else pos
            if pos >= 0:
                return len(string) - pos - 1
            else:
                return -1  # no point
    else:
        return -1  # not a number


def _padleft(width, s):
    """Flush right.

    >>> _padleft(6, '\u044f\u0439\u0446\u0430') == '  \u044f\u0439\u0446\u0430'
    True

    """
    fmt = "{0:>%ds}" % width
    return fmt.format(s)


def _padright(width, s):
    """Flush left.

    >>> _padright(6, '\u044f\u0439\u0446\u0430') == '\u044f\u0439\u0446\u0430  '
    True

    """
    fmt = "{0:<%ds}" % width
    return fmt.format(s)


def _padboth(width, s):
    """Center string.

    >>> _padboth(6, '\u044f\u0439\u0446\u0430') == ' \u044f\u0439\u0446\u0430 '
    True

    """
    fmt = "{0:^%ds}" % width
    return fmt.format(s)


def _padnone(ignore_width, s):
    return s


def _strip_invisible(s):
    "Remove invisible ANSI color codes."
    if isinstance(s, _text_type):
        return re.sub(_invisible_codes, "", s)
    else:  # a bytestring
        return re.sub(_invisible_codes_bytes, "", s)


def _visible_width(s):
    """Visible width of a printed string. ANSI color codes are removed.

    >>> _visible_width('\x1b[31mhello\x1b[0m'), _visible_width("world")
    (5, 5)

    """
    # optional wide-character support
    if wcwidth is not None and WIDE_CHARS_MODE:
        len_fn = wcwidth.wcswidth
    else:
        len_fn = len
    if isinstance(s, _text_type) or isinstance(s, _binary_type):
        return len_fn(_strip_invisible(s))
    else:
        return len_fn(_text_type(s))


def _is_multiline(s):
    if isinstance(s, _text_type):
        return bool(re.search(_multiline_codes, s))
    else:  # a bytestring
        return bool(re.search(_multiline_codes_bytes, s))


def _multiline_width(multiline_s, line_width_fn=len):
    """Visible width of a potentially multiline content."""
    return max(map(line_width_fn, re.split("[\r\n]", multiline_s)))


def _choose_width_fn(has_invisible, enable_widechars, is_multiline):
    """Return a function to calculate visible cell width."""
    if has_invisible:
        line_width_fn = _visible_width
    elif enable_widechars:  # optional wide-character support if available
        line_width_fn = wcwidth.wcswidth
    else:
        line_width_fn = len
    if is_multiline:
        width_fn = lambda s: _multiline_width(s, line_width_fn)  # noqa
    else:
        width_fn = line_width_fn
    return width_fn


def _align_column_choose_padfn(strings, alignment, has_invisible):
    if alignment == "right":
        if not PRESERVE_WHITESPACE:
            strings = [s.strip() for s in strings]
        padfn = _padleft
    elif alignment == "center":
        if not PRESERVE_WHITESPACE:
            strings = [s.strip() for s in strings]
        padfn = _padboth
    elif alignment == "decimal":
        if has_invisible:
            decimals = [_afterpoint(_strip_invisible(s)) for s in strings]
        else:
            decimals = [_afterpoint(s) for s in strings]
        maxdecimals = max(decimals)
        strings = [s + (maxdecimals - decs) * " " for s, decs in zip(strings, decimals)]
        padfn = _padleft
    elif not alignment:
        padfn = _padnone
    else:
        if not PRESERVE_WHITESPACE:
            strings = [s.strip() for s in strings]
        padfn = _padright
    return strings, padfn


def _align_column(
    strings,
    alignment,
    minwidth=0,
    has_invisible=True,
    enable_widechars=False,
    is_multiline=False,
):
    """[string] -> [padded_string]"""
    strings, padfn = _align_column_choose_padfn(strings, alignment, has_invisible)
    width_fn = _choose_width_fn(has_invisible, enable_widechars, is_multiline)

    s_widths = list(map(width_fn, strings))
    maxwidth = max(max(s_widths), minwidth)
    # TODO: refactor column alignment in single-line and multiline modes
    if is_multiline:
        if not enable_widechars and not has_invisible:
            padded_strings = [
                "\n".join([padfn(maxwidth, s) for s in ms.splitlines()])
                for ms in strings
            ]
        else:
            # enable wide-character width corrections
            s_lens = [max((len(s) for s in re.split("[\r\n]", ms))) for ms in strings]
            visible_widths = [maxwidth - (w - l) for w, l in zip(s_widths, s_lens)]
            # wcswidth and _visible_width don't count invisible characters;
            # padfn doesn't need to apply another correction
            padded_strings = [
                "\n".join([padfn(w, s) for s in (ms.splitlines() or ms)])
                for ms, w in zip(strings, visible_widths)
            ]
    else:  # single-line cell values
        if not enable_widechars and not has_invisible:
            padded_strings = [padfn(maxwidth, s) for s in strings]
        else:
            # enable wide-character width corrections
            s_lens = list(map(len, strings))
            visible_widths = [maxwidth - (w - l) for w, l in zip(s_widths, s_lens)]
            # wcswidth and _visible_width don't count invisible characters;
            # padfn doesn't need to apply another correction
            padded_strings = [padfn(w, s) for s, w in zip(strings, visible_widths)]
    return padded_strings


def _more_generic(type1, type2):
    types = {
        _none_type: 0,
        _bool_type: 1,
        int: 2,
        float: 3,
        _binary_type: 4,
        _text_type: 5,
    }
    invtypes = {
        5: _text_type,
        4: _binary_type,
        3: float,
        2: int,
        1: _bool_type,
        0: _none_type,
    }
    moregeneric = max(types.get(type1, 5), types.get(type2, 5))
    return invtypes[moregeneric]


def _column_type(strings, has_invisible=True, numparse=True):
    """The least generic type all column values are convertible to.

    >>> _column_type([True, False]) is _bool_type
    True
    >>> _column_type(["1", "2"]) is _int_type
    True
    >>> _column_type(["1", "2.3"]) is _float_type
    True
    >>> _column_type(["1", "2.3", "four"]) is _text_type
    True
    >>> _column_type(["four", '\u043f\u044f\u0442\u044c']) is _text_type
    True
    >>> _column_type([None, "brux"]) is _text_type
    True
    >>> _column_type([1, 2, None]) is _int_type
    True
    >>> import datetime as dt
    >>> _column_type([dt.datetime(1991,2,19), dt.time(17,35)]) is _text_type
    True

    """
    types = [_type(s, has_invisible, numparse) for s in strings]
    return reduce(_more_generic, types, _bool_type)


def _format(val, valtype, floatfmt, missingval="", has_invisible=True):
    """Format a value accoding to its type.

    Unicode is supported:

    >>> hrow = ['\u0431\u0443\u043a\u0432\u0430', '\u0446\u0438\u0444\u0440\u0430'] ; \
        tbl = [['\u0430\u0437', 2], ['\u0431\u0443\u043a\u0438', 4]] ; \
        good_result = '\\u0431\\u0443\\u043a\\u0432\\u0430      \\u0446\\u0438\\u0444\\u0440\\u0430\\n-------  -------\\n\\u0430\\u0437             2\\n\\u0431\\u0443\\u043a\\u0438           4' ; \
        tabulate(tbl, headers=hrow) == good_result
    True

    """  # noqa
    if val is None:
        return missingval

    if valtype in [int, _text_type]:
        return "{0}".format(val)
    elif valtype is _binary_type:
        try:
            return _text_type(val, "ascii")
        except TypeError:
            return _text_type(val)
    elif valtype is float:
        is_a_colored_number = has_invisible and isinstance(
            val, (_text_type, _binary_type)
        )
        if is_a_colored_number:
            raw_val = _strip_invisible(val)
            formatted_val = format(float(raw_val), floatfmt)
            return val.replace(raw_val, formatted_val)
        else:
            return format(float(val), floatfmt)
    else:
        return "{0}".format(val)


def _align_header(
    header, alignment, width, visible_width, is_multiline=False, width_fn=None
):
    "Pad string header to width chars given known visible_width of the header."
    if is_multiline:
        header_lines = re.split(_multiline_codes, header)
        padded_lines = [
            _align_header(h, alignment, width, width_fn(h)) for h in header_lines
        ]
        return "\n".join(padded_lines)
    # else: not multiline
    ninvisible = len(header) - visible_width
    width += ninvisible
    if alignment == "left":
        return _padright(width, header)
    elif alignment == "center":
        return _padboth(width, header)
    elif not alignment:
        return "{0}".format(header)
    else:
        return _padleft(width, header)


def _prepend_row_index(rows, index):
    """Add a left-most index column."""
    if index is None or index is False:
        return rows
    if len(index) != len(rows):
        print("index=", index)
        print("rows=", rows)
        raise ValueError("index must be as long as the number of data rows")
    rows = [[v] + list(row) for v, row in zip(index, rows)]
    return rows


def _bool(val):
    "A wrapper around standard bool() which doesn't throw on NumPy arrays"
    try:
        return bool(val)
    except ValueError:  # val is likely to be a numpy array with many elements
        return False


def _normalize_tabular_data(tabular_data, headers, showindex="default"):
    """Transform a supported data type to a list of lists, and a list of headers.

    Supported tabular data types:

    * list-of-lists or another iterable of iterables

    * list of named tuples (usually used with headers="keys")

    * list of dicts (usually used with headers="keys")

    * list of OrderedDicts (usually used with headers="keys")

    * 2D NumPy arrays

    * NumPy record arrays (usually used with headers="keys")

    * dict of iterables (usually used with headers="keys")

    * pandas.DataFrame (usually used with headers="keys")

    The first row can be used as headers if headers="firstrow",
    column indices can be used as headers if headers="keys".

    If showindex="default", show row indices of the pandas.DataFrame.
    If showindex="always", show row indices for all types of data.
    If showindex="never", don't show row indices for all types of data.
    If showindex is an iterable, show its values as row indices.

    """

    try:
        bool(headers)
        is_headers2bool_broken = False  # noqa
    except ValueError:  # numpy.ndarray, pandas.core.index.Index, ...
        is_headers2bool_broken = True  # noqa
        headers = list(headers)

    index = None
    if hasattr(tabular_data, "keys") and hasattr(tabular_data, "values"):
        # dict-like and pandas.DataFrame?
        if hasattr(tabular_data.values, "__call__"):
            # likely a conventional dict
            keys = tabular_data.keys()
            rows = list(
                izip_longest(*tabular_data.values())
            )  # columns have to be transposed
        elif hasattr(tabular_data, "index"):
            # values is a property, has .index => it's likely a pandas.DataFrame (pandas 0.11.0)
            keys = list(tabular_data)
            if tabular_data.index.name is not None:
                if isinstance(tabular_data.index.name, list):
                    keys[:0] = tabular_data.index.name
                else:
                    keys[:0] = [tabular_data.index.name]
            vals = tabular_data.values  # values matrix doesn't need to be transposed
            # for DataFrames add an index per default
            index = list(tabular_data.index)
            rows = [list(row) for row in vals]
        else:
            raise ValueError("tabular data doesn't appear to be a dict or a DataFrame")

        if headers == "keys":
            headers = list(map(_text_type, keys))  # headers should be strings

    else:  # it's a usual an iterable of iterables, or a NumPy array
        rows = list(tabular_data)

        if headers == "keys" and not rows:
            # an empty table (issue #81)
            headers = []
        elif (
            headers == "keys"
            and hasattr(tabular_data, "dtype")
            and getattr(tabular_data.dtype, "names")
        ):
            # numpy record array
            headers = tabular_data.dtype.names
        elif (
            headers == "keys"
            and len(rows) > 0
            and isinstance(rows[0], tuple)
            and hasattr(rows[0], "_fields")
        ):
            # namedtuple
            headers = list(map(_text_type, rows[0]._fields))
        elif len(rows) > 0 and isinstance(rows[0], dict):
            # dict or OrderedDict
            uniq_keys = set()  # implements hashed lookup
            keys = []  # storage for set
            if headers == "firstrow":
                firstdict = rows[0] if len(rows) > 0 else {}
                keys.extend(firstdict.keys())
                uniq_keys.update(keys)
                rows = rows[1:]
            for row in rows:
                for k in row.keys():
                    # Save unique items in input order
                    if k not in uniq_keys:
                        keys.append(k)
                        uniq_keys.add(k)
            if headers == "keys":
                headers = keys
            elif isinstance(headers, dict):
                # a dict of headers for a list of dicts
                headers = [headers.get(k, k) for k in keys]
                headers = list(map(_text_type, headers))
            elif headers == "firstrow":
                if len(rows) > 0:
                    headers = [firstdict.get(k, k) for k in keys]
                    headers = list(map(_text_type, headers))
                else:
                    headers = []
            elif headers:
                raise ValueError(
                    "headers for a list of dicts is not a dict or a keyword"
                )
            rows = [[row.get(k) for k in keys] for row in rows]

        elif (
            headers == "keys"
            and hasattr(tabular_data, "description")
            and hasattr(tabular_data, "fetchone")
            and hasattr(tabular_data, "rowcount")
        ):
            # Python Database API cursor object (PEP 0249)
            # print tabulate(cursor, headers='keys')
            headers = [column[0] for column in tabular_data.description]

        elif headers == "keys" and len(rows) > 0:
            # keys are column indices
            headers = list(map(_text_type, range(len(rows[0]))))

    # take headers from the first row if necessary
    if headers == "firstrow" and len(rows) > 0:
        if index is not None:
            headers = [index[0]] + list(rows[0])
            index = index[1:]
        else:
            headers = rows[0]
        headers = list(map(_text_type, headers))  # headers should be strings
        rows = rows[1:]

    headers = list(map(_text_type, headers))
    rows = list(map(list, rows))

    # add or remove an index column
    showindex_is_a_str = type(showindex) in [_text_type, _binary_type]
    if showindex == "default" and index is not None:
        rows = _prepend_row_index(rows, index)
    elif isinstance(showindex, Iterable) and not showindex_is_a_str:
        rows = _prepend_row_index(rows, list(showindex))
    elif showindex == "always" or (_bool(showindex) and not showindex_is_a_str):
        if index is None:
            index = list(range(len(rows)))
        rows = _prepend_row_index(rows, index)
    elif showindex == "never" or (not _bool(showindex) and not showindex_is_a_str):
        pass

    # pad with empty headers for initial columns if necessary
    if headers and len(rows) > 0:
        nhs = len(headers)
        ncols = len(rows[0])
        if nhs < ncols:
            headers = [""] * (ncols - nhs) + headers

    return rows, headers


def tabulate(
    tabular_data,
    headers=(),
    tablefmt="simple",
    floatfmt=_DEFAULT_FLOATFMT,
    numalign="decimal",
    stralign="left",
    missingval=_DEFAULT_MISSINGVAL,
    showindex="default",
    disable_numparse=False,
    colalign=None,
):
    """Format a fixed width table for pretty printing.

    >>> print(tabulate([[1, 2.34], [-56, "8.999"], ["2", "10001"]]))
    ---  ---------
      1      2.34
    -56      8.999
      2  10001
    ---  ---------

    The first required argument (`tabular_data`) can be a
    list-of-lists (or another iterable of iterables), a list of named
    tuples, a dictionary of iterables, an iterable of dictionaries,
    a two-dimensional NumPy array, NumPy record array, or a Pandas'
    dataframe.


    Table headers
    -------------

    To print nice column headers, supply the second argument (`headers`):

      - `headers` can be an explicit list of column headers
      - if `headers="firstrow"`, then the first row of data is used
      - if `headers="keys"`, then dictionary keys or column indices are used

    Otherwise a headerless table is produced.

    If the number of headers is less than the number of columns, they
    are supposed to be names of the last columns. This is consistent
    with the plain-text format of R and Pandas' dataframes.

    >>> print(tabulate([["sex","age"],["Alice","F",24],["Bob","M",19]],
    ...       headers="firstrow"))
           sex      age
    -----  -----  -----
    Alice  F         24
    Bob    M         19

    By default, pandas.DataFrame data have an additional column called
    row index. To add a similar column to all other types of data,
    use `showindex="always"` or `showindex=True`. To suppress row indices
    for all types of data, pass `showindex="never" or `showindex=False`.
    To add a custom row index column, pass `showindex=some_iterable`.

    >>> print(tabulate([["F",24],["M",19]], showindex="always"))
    -  -  --
    0  F  24
    1  M  19
    -  -  --


    Column alignment
    ----------------

    `tabulate` tries to detect column types automatically, and aligns
    the values properly. By default it aligns decimal points of the
    numbers (or flushes integer numbers to the right), and flushes
    everything else to the left. Possible column alignments
    (`numalign`, `stralign`) are: "right", "center", "left", "decimal"
    (only for `numalign`), and None (to disable alignment).


    Table formats
    -------------

    `floatfmt` is a format specification used for columns which
    contain numeric data with a decimal point. This can also be
    a list or tuple of format strings, one per column.

    `None` values are replaced with a `missingval` string (like
    `floatfmt`, this can also be a list of values for different
    columns):

    >>> print(tabulate([["spam", 1, None],
    ...                 ["eggs", 42, 3.14],
    ...                 ["other", None, 2.7]], missingval="?"))
    -----  --  ----
    spam    1  ?
    eggs   42  3.14
    other   ?  2.7
    -----  --  ----

    Various plain-text table formats (`tablefmt`) are supported:
    'plain', 'simple', 'grid', 'pipe', 'orgtbl', 'rst', 'mediawiki',
    'latex', 'latex_raw' and 'latex_booktabs'. Variable `tabulate_formats`
    contains the list of currently supported formats.

    "plain" format doesn't use any pseudographics to draw tables,
    it separates columns with a double space:

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]],
    ...                 ["strings", "numbers"], "plain"))
    strings      numbers
    spam         41.9999
    eggs        451

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]], tablefmt="plain"))
    spam   41.9999
    eggs  451

    "simple" format is like Pandoc simple_tables:

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]],
    ...                 ["strings", "numbers"], "simple"))
    strings      numbers
    ---------  ---------
    spam         41.9999
    eggs        451

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]], tablefmt="simple"))
    ----  --------
    spam   41.9999
    eggs  451
    ----  --------

    "grid" is similar to tables produced by Emacs table.el package or
    Pandoc grid_tables:

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]],
    ...                ["strings", "numbers"], "grid"))
    +-----------+-----------+
    | strings   |   numbers |
    +===========+===========+
    | spam      |   41.9999 |
    +-----------+-----------+
    | eggs      |  451      |
    +-----------+-----------+

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]], tablefmt="grid"))
    +------+----------+
    | spam |  41.9999 |
    +------+----------+
    | eggs | 451      |
    +------+----------+

    "fancy_grid" draws a grid using box-drawing characters:

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]],
    ...                ["strings", "numbers"], "fancy_grid"))
    ╒═══════════╤═══════════╕
    │ strings   │   numbers │
    ╞═══════════╪═══════════╡
    │ spam      │   41.9999 │
    ├───────────┼───────────┤
    │ eggs      │  451      │
    ╘═══════════╧═══════════╛

    "pipe" is like tables in PHP Markdown Extra extension or Pandoc
    pipe_tables:

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]],
    ...                ["strings", "numbers"], "pipe"))
    | strings   |   numbers |
    |:----------|----------:|
    | spam      |   41.9999 |
    | eggs      |  451      |

    "presto" is like tables produce by the Presto CLI:

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]],
    ...                ["strings", "numbers"], "presto"))
     strings   |   numbers
    -----------+-----------
     spam      |   41.9999
     eggs      |  451

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]], tablefmt="pipe"))
    |:-----|---------:|
    | spam |  41.9999 |
    | eggs | 451      |

    "orgtbl" is like tables in Emacs org-mode and orgtbl-mode. They
    are slightly different from "pipe" format by not using colons to
    define column alignment, and using a "+" sign to indicate line
    intersections:

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]],
    ...                ["strings", "numbers"], "orgtbl"))
    | strings   |   numbers |
    |-----------+-----------|
    | spam      |   41.9999 |
    | eggs      |  451      |


    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]], tablefmt="orgtbl"))
    | spam |  41.9999 |
    | eggs | 451      |

    "rst" is like a simple table format from reStructuredText; please
    note that reStructuredText accepts also "grid" tables:

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]],
    ...                ["strings", "numbers"], "rst"))
    =========  =========
    strings      numbers
    =========  =========
    spam         41.9999
    eggs        451
    =========  =========

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]], tablefmt="rst"))
    ====  ========
    spam   41.9999
    eggs  451
    ====  ========

    "mediawiki" produces a table markup used in Wikipedia and on other
    MediaWiki-based sites:

    >>> print(tabulate([["strings", "numbers"], ["spam", 41.9999], ["eggs", "451.0"]],
    ...                headers="firstrow", tablefmt="mediawiki"))
    {| class="wikitable" style="text-align: left;"
    |+ <!-- caption -->
    |-
    ! strings   !! align="right"|   numbers
    |-
    | spam      || align="right"|   41.9999
    |-
    | eggs      || align="right"|  451
    |}

    "html" produces HTML markup as an html.escape'd str
    with a ._repr_html_ method so that Jupyter Lab and Notebook display the HTML
    and a .str property so that the raw HTML remains accessible:

    >>> print(tabulate([["strings", "numbers"], ["spam", 41.9999], ["eggs", "451.0"]],
    ...                headers="firstrow", tablefmt="html"))
    <table>
    <thead>
    <tr><th>strings  </th><th style="text-align: right;">  numbers</th></tr>
    </thead>
    <tbody>
    <tr><td>spam     </td><td style="text-align: right;">  41.9999</td></tr>
    <tr><td>eggs     </td><td style="text-align: right;"> 451     </td></tr>
    </tbody>
    </table>

    "latex" produces a tabular environment of LaTeX document markup:

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]], tablefmt="latex"))
    \\begin{tabular}{lr}
    \\hline
     spam &  41.9999 \\\\
     eggs & 451      \\\\
    \\hline
    \\end{tabular}

    "latex_raw" is similar to "latex", but doesn't escape special characters,
    such as backslash and underscore, so LaTeX commands may embedded into
    cells' values:

    >>> print(tabulate([["spam$_9$", 41.9999], ["\\\\emph{eggs}", "451.0"]], tablefmt="latex_raw"))
    \\begin{tabular}{lr}
    \\hline
     spam$_9$    &  41.9999 \\\\
     \\emph{eggs} & 451      \\\\
    \\hline
    \\end{tabular}

    "latex_booktabs" produces a tabular environment of LaTeX document markup
    using the booktabs.sty package:

    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]], tablefmt="latex_booktabs"))
    \\begin{tabular}{lr}
    \\toprule
     spam &  41.9999 \\\\
     eggs & 451      \\\\
    \\bottomrule
    \\end{tabular}

    Number parsing
    --------------
    By default, anything which can be parsed as a number is a number.
    This ensures numbers represented as strings are aligned properly.
    This can lead to weird results for particular strings such as
    specific git SHAs e.g. "42992e1" will be parsed into the number
    429920 and aligned as such.

    To completely disable number parsing (and alignment), use
    `disable_numparse=True`. For more fine grained control, a list column
    indices is used to disable number parsing only on those columns
    e.g. `disable_numparse=[0, 2]` would disable number parsing only on the
    first and third columns.
    """
    if tabular_data is None:
        tabular_data = []
    list_of_lists, headers = _normalize_tabular_data(
        tabular_data, headers, showindex=showindex
    )

    # empty values in the first column of RST tables should be escaped (issue #82)
    # "" should be escaped as "\\ " or ".."
    if tablefmt == "rst":
        list_of_lists, headers = _rst_escape_first_column(list_of_lists, headers)

    # PrettyTable formatting does not use any extra padding.
    # Numbers are not parsed and are treated the same as strings for alignment.
    # Check if pretty is the format being used and override the defaults so it
    # does not impact other formats.
    min_padding = MIN_PADDING
    if tablefmt == "pretty":
        min_padding = 0
        disable_numparse = True
        numalign = "center"
        stralign = "center"

    # optimization: look for ANSI control codes once,
    # enable smart width functions only if a control code is found
    plain_text = "\t".join(
        ["\t".join(map(_text_type, headers))]
        + ["\t".join(map(_text_type, row)) for row in list_of_lists]
    )

    has_invisible = re.search(_invisible_codes, plain_text)
    enable_widechars = wcwidth is not None and WIDE_CHARS_MODE
    if (
        not isinstance(tablefmt, TableFormat)
        and tablefmt in multiline_formats
        and _is_multiline(plain_text)
    ):
        tablefmt = multiline_formats.get(tablefmt, tablefmt)
        is_multiline = True
    else:
        is_multiline = False
    width_fn = _choose_width_fn(has_invisible, enable_widechars, is_multiline)

    # format rows and columns, convert numeric values to strings
    cols = list(izip_longest(*list_of_lists))
    numparses = _expand_numparse(disable_numparse, len(cols))
    coltypes = [_column_type(col, numparse=np) for col, np in zip(cols, numparses)]
    if isinstance(floatfmt, basestring):  # old version
        float_formats = len(cols) * [
            floatfmt
        ]  # just duplicate the string to use in each column
    else:  # if floatfmt is list, tuple etc we have one per column
        float_formats = list(floatfmt)
        if len(float_formats) < len(cols):
            float_formats.extend((len(cols) - len(float_formats)) * [_DEFAULT_FLOATFMT])
    if isinstance(missingval, basestring):
        missing_vals = len(cols) * [missingval]
    else:
        missing_vals = list(missingval)
        if len(missing_vals) < len(cols):
            missing_vals.extend((len(cols) - len(missing_vals)) * [_DEFAULT_MISSINGVAL])
    cols = [
        [_format(v, ct, fl_fmt, miss_v, has_invisible) for v in c]
        for c, ct, fl_fmt, miss_v in zip(cols, coltypes, float_formats, missing_vals)
    ]

    # align columns
    aligns = [numalign if ct in [int, float] else stralign for ct in coltypes]
    if colalign is not None:
        assert isinstance(colalign, Iterable)
        for idx, align in enumerate(colalign):
            aligns[idx] = align
    minwidths = (
        [width_fn(h) + min_padding for h in headers] if headers else [0] * len(cols)
    )
    cols = [
        _align_column(c, a, minw, has_invisible, enable_widechars, is_multiline)
        for c, a, minw in zip(cols, aligns, minwidths)
    ]

    if headers:
        # align headers and add headers
        t_cols = cols or [[""]] * len(headers)
        t_aligns = aligns or [stralign] * len(headers)
        minwidths = [
            max(minw, max(width_fn(cl) for cl in c))
            for minw, c in zip(minwidths, t_cols)
        ]
        headers = [
            _align_header(h, a, minw, width_fn(h), is_multiline, width_fn)
            for h, a, minw in zip(headers, t_aligns, minwidths)
        ]
        rows = list(zip(*cols))
    else:
        minwidths = [max(width_fn(cl) for cl in c) for c in cols]
        rows = list(zip(*cols))

    if not isinstance(tablefmt, TableFormat):
        tablefmt = _table_formats.get(tablefmt, _table_formats["simple"])

    return _format_table(tablefmt, headers, rows, minwidths, aligns, is_multiline)


def _expand_numparse(disable_numparse, column_count):
    """
    Return a list of bools of length `column_count` which indicates whether
    number parsing should be used on each column.
    If `disable_numparse` is a list of indices, each of those indices are False,
    and everything else is True.
    If `disable_numparse` is a bool, then the returned list is all the same.
    """
    if isinstance(disable_numparse, Iterable):
        numparses = [True] * column_count
        for index in disable_numparse:
            numparses[index] = False
        return numparses
    else:
        return [not disable_numparse] * column_count


def _pad_row(cells, padding):
    if cells:
        pad = " " * padding
        padded_cells = [pad + cell + pad for cell in cells]
        return padded_cells
    else:
        return cells


def _build_simple_row(padded_cells, rowfmt):
    "Format row according to DataRow format without padding."
    begin, sep, end = rowfmt
    return (begin + sep.join(padded_cells) + end).rstrip()


def _build_row(padded_cells, colwidths, colaligns, rowfmt):
    "Return a string which represents a row of data cells."
    if not rowfmt:
        return None
    if hasattr(rowfmt, "__call__"):
        return rowfmt(padded_cells, colwidths, colaligns)
    else:
        return _build_simple_row(padded_cells, rowfmt)


def _append_basic_row(lines, padded_cells, colwidths, colaligns, rowfmt):
    lines.append(_build_row(padded_cells, colwidths, colaligns, rowfmt))
    return lines


def _append_multiline_row(
    lines, padded_multiline_cells, padded_widths, colaligns, rowfmt, pad
):
    colwidths = [w - 2 * pad for w in padded_widths]
    cells_lines = [c.splitlines() for c in padded_multiline_cells]
    nlines = max(map(len, cells_lines))  # number of lines in the row
    # vertically pad cells where some lines are missing
    cells_lines = [
        (cl + [" " * w] * (nlines - len(cl))) for cl, w in zip(cells_lines, colwidths)
    ]
    lines_cells = [[cl[i] for cl in cells_lines] for i in range(nlines)]
    for ln in lines_cells:
        padded_ln = _pad_row(ln, pad)
        _append_basic_row(lines, padded_ln, colwidths, colaligns, rowfmt)
    return lines


def _build_line(colwidths, colaligns, linefmt):
    "Return a string which represents a horizontal line."
    if not linefmt:
        return None
    if hasattr(linefmt, "__call__"):
        return linefmt(colwidths, colaligns)
    else:
        begin, fill, sep, end = linefmt
        cells = [fill * w for w in colwidths]
        return _build_simple_row(cells, (begin, sep, end))


def _append_line(lines, colwidths, colaligns, linefmt):
    lines.append(_build_line(colwidths, colaligns, linefmt))
    return lines


class JupyterHTMLStr(str):
    """Wrap the string with a _repr_html_ method so that Jupyter
    displays the HTML table"""

    def _repr_html_(self):
        return self

    @property
    def str(self):
        """add a .str property so that the raw string is still accessible"""
        return self


def _format_table(fmt, headers, rows, colwidths, colaligns, is_multiline):
    """Produce a plain-text representation of the table."""
    lines = []
    hidden = fmt.with_header_hide if (headers and fmt.with_header_hide) else []
    pad = fmt.padding
    headerrow = fmt.headerrow

    padded_widths = [(w + 2 * pad) for w in colwidths]
    if is_multiline:
        pad_row = lambda row, _: row  # noqa do it later, in _append_multiline_row
        append_row = partial(_append_multiline_row, pad=pad)
    else:
        pad_row = _pad_row
        append_row = _append_basic_row

    padded_headers = pad_row(headers, pad)
    padded_rows = [pad_row(row, pad) for row in rows]

    if fmt.lineabove and "lineabove" not in hidden:
        _append_line(lines, padded_widths, colaligns, fmt.lineabove)

    if padded_headers:
        append_row(lines, padded_headers, padded_widths, colaligns, headerrow)
        if fmt.linebelowheader and "linebelowheader" not in hidden:
            _append_line(lines, padded_widths, colaligns, fmt.linebelowheader)

    if padded_rows and fmt.linebetweenrows and "linebetweenrows" not in hidden:
        # initial rows with a line below
        for row in padded_rows[:-1]:
            append_row(lines, row, padded_widths, colaligns, fmt.datarow)
            _append_line(lines, padded_widths, colaligns, fmt.linebetweenrows)
        # the last row without a line below
        append_row(lines, padded_rows[-1], padded_widths, colaligns, fmt.datarow)
    else:
        for row in padded_rows:
            append_row(lines, row, padded_widths, colaligns, fmt.datarow)

    if fmt.linebelow and "linebelow" not in hidden:
        _append_line(lines, padded_widths, colaligns, fmt.linebelow)

    if headers or rows:
        output = "\n".join(lines)
        if fmt.lineabove == _html_begin_table_without_header:
            return JupyterHTMLStr(output)
        else:
            return output
    else:  # a completely empty table
        return ""


def _main():
    """\
    Usage: tabulate [options] [FILE ...]

    Pretty-print tabular data.
    See also https://github.com/astanin/python-tabulate

    FILE                      a filename of the file with tabular data;
                              if "-" or missing, read data from stdin.

    Options:

    -h, --help                show this message
    -1, --header              use the first row of data as a table header
    -o FILE, --output FILE    print table to FILE (default: stdout)
    -s REGEXP, --sep REGEXP   use a custom column separator (default: whitespace)
    -F FPFMT, --float FPFMT   floating point number format (default: g)
    -f FMT, --format FMT      set output table format; supported formats:
                              plain, simple, grid, fancy_grid, pipe, orgtbl,
                              rst, mediawiki, html, latex, latex_raw,
                              latex_booktabs, tsv
                              (default: simple)
    """
    import getopt
    import sys
    import textwrap

    usage = textwrap.dedent(_main.__doc__)
    try:
        opts, args = getopt.getopt(
            sys.argv[1:],
            "h1o:s:F:A:f:",
            ["help", "header", "output", "sep=", "float=", "align=", "format="],
        )
    except getopt.GetoptError as e:
        print(e)
        print(usage)
        sys.exit(2)
    headers = []
    floatfmt = _DEFAULT_FLOATFMT
    colalign = None
    tablefmt = "simple"
    sep = r"\s+"
    outfile = "-"
    for opt, value in opts:
        if opt in ["-1", "--header"]:
            headers = "firstrow"
        elif opt in ["-o", "--output"]:
            outfile = value
        elif opt in ["-F", "--float"]:
            floatfmt = value
        elif opt in ["-C", "--colalign"]:
            colalign = value.split()
        elif opt in ["-f", "--format"]:
            if value not in tabulate_formats:
                print("%s is not a supported table format" % value)
                print(usage)
                sys.exit(3)
            tablefmt = value
        elif opt in ["-s", "--sep"]:
            sep = value
        elif opt in ["-h", "--help"]:
            print(usage)
            sys.exit(0)
    files = [sys.stdin] if not args else args
    with (sys.stdout if outfile == "-" else open(outfile, "w")) as out:
        for f in files:
            if f == "-":
                f = sys.stdin
            if _is_file(f):
                _pprint_file(
                    f,
                    headers=headers,
                    tablefmt=tablefmt,
                    sep=sep,
                    floatfmt=floatfmt,
                    file=out,
                    colalign=colalign,
                )
            else:
                with open(f) as fobj:
                    _pprint_file(
                        fobj,
                        headers=headers,
                        tablefmt=tablefmt,
                        sep=sep,
                        floatfmt=floatfmt,
                        file=out,
                        colalign=colalign,
                    )


def _pprint_file(fobject, headers, tablefmt, sep, floatfmt, file, colalign):
    rows = fobject.readlines()
    table = [re.split(sep, r.rstrip()) for r in rows if r.strip()]
    print(
        tabulate(table, headers, tablefmt, floatfmt=floatfmt, colalign=colalign),
        file=file,
    )


if __name__ == "__main__":
    _main()



from functools import partial
from math import ceil

import pytest
from conftest import TestProjectBase

import flow
from flow.aggregates import (
    _AggregateStoresCursor,
    _JobAggregateCursor,
    aggregator,
    get_aggregate_id,
)


def generate_flow_project():
    class SimpleAggregateProject(flow.FlowProject):
        pass

    @SimpleAggregateProject.operation(aggregator=aggregator.groupby("is_even"))
    def op1(job):
        pass

    @SimpleAggregateProject.operation(aggregator=aggregator.groupby("is_even"))
    def op2(job):
        pass

    @SimpleAggregateProject.operation(aggregator=aggregator.groupsof(2))
    def op3(job):
        pass

    @SimpleAggregateProject.operation(aggregator=aggregator())
    def op4(job):
        pass

    @SimpleAggregateProject.operation
    def op5(job):
        pass

    return SimpleAggregateProject


class AggregateProjectSetup(TestProjectBase):
    project_class = generate_flow_project()

    def mock_project(self):
        project = self.project_class.get_project(path=self._tmp_dir.name)
        for i in range(10):
            is_even = (i % 2) == 0
            if is_even:
                project.open_job(dict(i=i, half=i / 2, is_even=is_even)).init()
            else:
                project.open_job(dict(i=i, is_even=is_even)).init()
        project._reregister_aggregates()
        return project

    @pytest.fixture
    def mocked_project(self):
        return self.mock_project()


class AggregateFixtures:
    @pytest.fixture
    def get_single_job_aggregate(self):
        def generator(jobs):
            yield from ((job,) for job in jobs)

        return generator

    @pytest.fixture
    def get_all_job_aggregate(self):
        def generator(jobs):
            return (tuple(jobs),)

        return generator

    @pytest.fixture
    def list_of_aggregators(self, get_single_job_aggregate, get_all_job_aggregate):
        # The below list contains 14 distinct aggregator objects and some duplicates.
        return [
            aggregator(),
            aggregator(),
            aggregator(get_all_job_aggregate),
            aggregator(get_single_job_aggregate),
            aggregator(get_single_job_aggregate),
            aggregator.groupsof(1),
            aggregator.groupsof(1),
            aggregator.groupsof(2),
            aggregator.groupsof(3),
            aggregator.groupsof(4),
            aggregator.groupby("is_even"),
            aggregator.groupby("is_even"),
            aggregator.groupby("half", -1),
            aggregator.groupby("half", -1),
            aggregator.groupby(["half", "is_even"], default=[-1, -1]),
        ]

    def create_aggregate_store(self, aggregator_instance, mocked_project):
        return aggregator_instance._create_AggregateStore(mocked_project)

    def get_aggregates_from_store(self, aggregate_store):
        return tuple(aggregate_store.values())


# Test the _AggregateStore and _DefaultAggregateStore classes
class TestAggregateStore(AggregateProjectSetup, AggregateFixtures):
    def test_custom_aggregator_function(
        self, mocked_project, get_single_job_aggregate, get_all_job_aggregate
    ):
        # Testing aggregator function returning aggregates of 1
        aggregate_store = self.create_aggregate_store(
            aggregator(get_single_job_aggregate), mocked_project
        )
        aggregate_job_manual = tuple(get_single_job_aggregate(mocked_project))
        assert self.get_aggregates_from_store(aggregate_store) == aggregate_job_manual

        # Testing aggregator function returning aggregates of all the jobs
        aggregate_store = self.create_aggregate_store(
            aggregator(get_all_job_aggregate), mocked_project
        )
        assert self.get_aggregates_from_store(aggregate_store) == get_all_job_aggregate(
            mocked_project
        )

    def test_sort_by(self, mocked_project):
        helper_sort = partial(sorted, key=lambda job: job.sp.i)
        aggregate_store = self.create_aggregate_store(
            aggregator(sort_by="i"), mocked_project
        )
        assert self.get_aggregates_from_store(aggregate_store) == (
            tuple(helper_sort(mocked_project)),
        )

    def test_sort_by_callable(self, mocked_project):
        def keyfunction(job):
            return job.sp.i

        helper_sort = partial(sorted, key=keyfunction)
        aggregate_store = self.create_aggregate_store(
            aggregator(sort_by=keyfunction), mocked_project
        )
        assert self.get_aggregates_from_store(aggregate_store) == (
            tuple(helper_sort(mocked_project)),
        )

    def test_sort_descending(self, mocked_project):
        helper_sort = partial(sorted, key=lambda job: job.sp.i, reverse=True)
        aggregate_store = self.create_aggregate_store(
            aggregator(sort_by="i", sort_ascending=False), mocked_project
        )
        assert self.get_aggregates_from_store(aggregate_store) == (
            tuple(helper_sort(mocked_project)),
        )

    @pytest.mark.parametrize("aggregate_length", [1, 2, 3, 6, 10])
    def test_groups_of_valid_num(self, mocked_project, aggregate_length):
        aggregate_store = self.create_aggregate_store(
            aggregator.groupsof(aggregate_length), mocked_project
        )
        expected_length = ceil(10 / aggregate_length)
        assert len(aggregate_store) == expected_length

        # We also check the length of every aggregate in order to ensure
        # proper aggregation.
        last_agg_length = (
            10 % aggregate_length if (10 % aggregate_length != 0) else aggregate_length
        )
        aggregates = list(aggregate_store.values())
        assert all(len(agg) == aggregate_length for agg in aggregates[:-1])
        assert len(aggregates[-1]) == last_agg_length

    def test_groupby_with_valid_string_key(self, mocked_project):
        aggregate_store = self.create_aggregate_store(
            aggregator.groupby("is_even"), mocked_project
        )
        assert len(aggregate_store) == 2
        for aggregate in aggregate_store.values():
            even = aggregate[0].sp.is_even
            assert all(even == job.sp.is_even for job in aggregate)

    def test_groupby_with_invalid_string_key(self, mocked_project):
        with pytest.raises(KeyError):
            # We will attempt to generate aggregates but will fail in
            # doing so due to the invalid key
            self.create_aggregate_store(
                aggregator.groupby("invalid_key"), mocked_project
            )

    def test_groupby_with_default_key_for_string(self, mocked_project):
        aggregate_store = self.create_aggregate_store(
            aggregator.groupby("half", default=-1), mocked_project
        )
        assert len(aggregate_store) == 6
        for aggregate in aggregate_store.values():
            half = aggregate[0].sp.get("half", -1)
            assert all(half == job.sp.get("half", -1) for job in aggregate)

    def test_groupby_with_Iterable_key(self, mocked_project):
        aggregate_store = self.create_aggregate_store(
            aggregator.groupby(["i", "is_even"]), mocked_project
        )
        # No aggregation takes place hence this means we don't need to check
        # whether all the aggregate members are equivalent.
        assert len(aggregate_store) == len(mocked_project)

    def test_groupby_with_invalid_Iterable_key(self, mocked_project):
        with pytest.raises(KeyError):
            # We will attempt to generate aggregates but will fail in
            # doing so due to the invalid keys
            self.create_aggregate_store(
                aggregator.groupby(["half", "is_even"]), mocked_project
            )

    def test_groupby_with_valid_default_key_for_Iterable(self, mocked_project):
        aggregate_store = self.create_aggregate_store(
            aggregator.groupby(["half", "is_even"], default=[-1, -1]), mocked_project
        )
        assert len(aggregate_store) == 6
        for aggregate in aggregate_store.values():
            half = aggregate[0].sp.get("half", -1)
            even = aggregate[0].sp["is_even"]
            assert all(
                half == job.sp.get("half", -1) and even == job.sp["is_even"]
                for job in aggregate
            )

    def test_groupby_with_callable_key(self, mocked_project):
        def keyfunction(job):
            return job.sp["is_even"]

        aggregate_store = self.create_aggregate_store(
            aggregator.groupby(keyfunction), mocked_project
        )
        assert len(aggregate_store) == 2
        for aggregate in aggregate_store.values():
            even = aggregate[0].sp.is_even
            assert all(even == job.sp.is_even for job in aggregate)

    def test_groupby_with_invalid_callable_key(self, mocked_project):
        def keyfunction(job):
            return job.sp["half"]

        with pytest.raises(KeyError):
            # We will attempt to generate aggregates but will fail in
            # doing so due to the invalid key
            self.create_aggregate_store(aggregator.groupby(keyfunction), mocked_project)

    def test_valid_select(self, mocked_project):
        def select(job):
            return job.sp.i > 5

        aggregate_store = self.create_aggregate_store(
            aggregator.groupsof(1, select=select), mocked_project
        )
        selected_jobs = tuple((job,) for job in mocked_project if select(job))
        assert self.get_aggregates_from_store(aggregate_store) == selected_jobs

    def test_store_hashing(self, mocked_project, list_of_aggregators):
        # Since we need to store groups on a per aggregate basis in the mocked_project,
        # we need to be sure that the aggregates are hashing and compared correctly.
        list_of_stores = [
            self.create_aggregate_store(aggregator, mocked_project)
            for aggregator in list_of_aggregators
        ]
        # The above list contains 14 distinct store objects (because a
        # store object is differentiated on the basis of the
        # ``_is_default_aggregate`` attribute of the aggregator). When this
        # list is converted to a set, then these objects are hashed first and
        # then compared. Since sets don't carry duplicate values, we test
        # whether the length of the set obtained from the list is equal to 14
        # or not.
        assert len(set(list_of_stores)) == 14

    @pytest.mark.parametrize(
        "aggregator_instance",
        [
            aggregator(),
            aggregator.groupsof(1),
            aggregator.groupsof(2),
            aggregator.groupsof(3),
            aggregator.groupsof(4),
            aggregator.groupby("is_even"),
            aggregator.groupby("half", -1),
            aggregator.groupby(["half", "is_even"], default=[-1, -1]),
        ],
    )
    def test_get_by_id(self, mocked_project, aggregator_instance):
        # Ensure that all aggregates can be fetched by id.
        aggregate_store = self.create_aggregate_store(
            aggregator_instance, mocked_project
        )
        for aggregate in aggregate_store.values():
            assert aggregate == aggregate_store[get_aggregate_id(aggregate)]

    def test_get_invalid_id(self, mocked_project):
        jobs = tuple(mocked_project)
        full_aggregate_store = self.create_aggregate_store(aggregator(), mocked_project)
        default_aggregate_store = self.create_aggregate_store(
            aggregator.groupsof(1), mocked_project
        )
        # Test for an aggregate of single job for aggregator instance
        with pytest.raises(LookupError):
            full_aggregate_store[get_aggregate_id((jobs[0],))]
        # Test for an aggregate of all jobs for default aggregator
        with pytest.raises(LookupError):
            default_aggregate_store[get_aggregate_id(jobs)]

    def test_contains(self, mocked_project):
        jobs = tuple(mocked_project)
        full_aggregate_store = self.create_aggregate_store(aggregator(), mocked_project)
        default_aggregate_store = aggregator.groupsof(1)._create_AggregateStore(
            mocked_project
        )
        # Test for an aggregate of all jobs
        assert get_aggregate_id(jobs) in full_aggregate_store
        assert get_aggregate_id(jobs) not in default_aggregate_store
        # Test for an aggregate of single job
        assert not jobs[0].id in full_aggregate_store
        assert jobs[0].id in default_aggregate_store


# Test the decorator class aggregator
class TestAggregate(AggregateProjectSetup, AggregateFixtures):
    def test_default_init(self):
        aggregate_instance = aggregator()
        assert not aggregate_instance._is_default_aggregator
        assert aggregate_instance._sort_by is None
        assert aggregate_instance._sort_ascending
        assert aggregate_instance._select is None
        assert next(aggregate_instance._aggregator_function((1, 2, 3))) == (1, 2, 3)

    @pytest.mark.parametrize("aggregator_function", ["str", 1, {}])
    def test_invalid_aggregator_function(self, aggregator_function):
        with pytest.raises(TypeError):
            aggregator(aggregator_function)

    @pytest.mark.parametrize("sort_by", [1, {}])
    def test_invalid_sort_by(self, sort_by):
        with pytest.raises(TypeError):
            aggregator(sort_by=sort_by)

    @pytest.mark.parametrize("select", ["str", 1, []])
    def test_invalid_select(self, select):
        with pytest.raises(TypeError):
            aggregator(select=select)

    @pytest.mark.parametrize("invalid_value", [{}, "str", -1, -1.5])
    def test_groups_of_invalid_num(self, invalid_value):
        with pytest.raises((TypeError, ValueError)):
            aggregator.groupsof(invalid_value)

    def test_group_by_invalid_key(self):
        with pytest.raises(TypeError):
            aggregator.groupby(1)

    def test_groupby_with_valid_type_default_for_Iterable(self):
        aggregator.groupby(["half", "is_even"], default=[-1, -1])

    def test_groupby_with_invalid_type_default_key_for_Iterable(self):
        with pytest.raises(TypeError):
            aggregator.groupby(["half", "is_even"], default=-1)

    def test_groupby_with_invalid_length_default_key_for_Iterable(self):
        with pytest.raises(ValueError):
            aggregator.groupby(["half", "is_even"], default=[-1, -1, -1])

    def test_aggregate_hashing(self, list_of_aggregators):
        # Since we need to store groups on a per aggregate basis in the project,
        # we need to be sure that the aggregates are hashing and compared correctly.
        # This test ensures this feature.
        # list_of_aggregators contains 14 distinct store objects (because an
        # aggregator object is differentiated on the basis of the `_is_aggregate` attribute).
        # When this list is converted to set, then these objects are hashed first
        # and then compared. Since sets don't carry duplicate values, we test
        # whether the length of the set obtained from the list is equal to 14 or not.
        assert len(set(list_of_aggregators)) == 14
        # Ensure that equality implies hash equality.
        for agg1 in list_of_aggregators:
            for agg2 in list_of_aggregators:
                if agg1 == agg2:
                    assert hash(agg1) == hash(agg2)


class TestAggregateUtilities(AggregateProjectSetup):
    def test_skips_duplicate_aggregates(self, mocked_project):
        # Check that length and actual iteration are the same and correct.
        # If this did not skip duplicates the length would be 19.
        stores = _AggregateStoresCursor(mocked_project)
        assert len(stores) == 18
        assert len(list(stores)) == 18
        # Ensure that some known aggregates are in the cursor
        assert tuple(mocked_project) in stores
        assert all((job,) in stores for job in mocked_project)

    def test_filters(self, mocked_project):
        agg_cursor = _JobAggregateCursor(
            project=mocked_project, filter={"is_even": True}
        )
        assert len(agg_cursor) == 5

    def test_reregister_aggregates(self, mocked_project):
        agg_cursor = _AggregateStoresCursor(project=mocked_project)
        NUM_BEFORE_REREGISTRATION = 18
        assert len(agg_cursor) == NUM_BEFORE_REREGISTRATION
        new_jobs = [
            mocked_project.open_job(dict(i=10, is_even=True, half=5)),
            mocked_project.open_job(dict(i=11, is_even=False)),
        ]
        assert not any(j in mocked_project for j in new_jobs)
        for job in new_jobs:
            job.init()
        # Default aggregate store doesn't need to be re-registered.
        assert len(agg_cursor) == NUM_BEFORE_REREGISTRATION + 2
        mocked_project._reregister_aggregates()
        # The operation op3 adds another aggregate in the mocked_project.
        assert len(agg_cursor) == NUM_BEFORE_REREGISTRATION + 3
